import re
import warnings
from typing import Literal, overload

from cognite.neat.exceptions import wrangle_warnings
from cognite.neat.legacy.rules import exceptions
from cognite.neat.legacy.rules.models.rules import (
    Rules,
    dms_property_id_compliance_regex,
    value_id_compliance_regex,
    view_id_compliance_regex,
)


@overload
def are_entity_names_dms_compliant(
    transformation_rules: Rules, return_report: Literal[True]
) -> tuple[bool, list[dict]]: ...


@overload
def are_entity_names_dms_compliant(transformation_rules: Rules, return_report: Literal[False] = False) -> bool: ...


def are_entity_names_dms_compliant(
    transformation_rules: Rules, return_report: bool = False
) -> bool | tuple[bool, list[dict]]:
    """Check if data model definitions are valid."""

    flag: bool = True
    with warnings.catch_warnings(record=True) as validation_warnings:
        for class_ in transformation_rules.classes.values():
            if not re.match(view_id_compliance_regex, class_.class_id):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Class", class_.class_id, f"[Classes/Class/{class_.class_id}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

        for row, property_ in transformation_rules.properties.items():
            # check class id which would resolve as view/container id
            if not re.match(view_id_compliance_regex, property_.class_id):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Class", property_.class_id, f"[Properties/Class/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

            # check property id which would resolve as view/container id
            if not re.match(dms_property_id_compliance_regex, property_.property_id):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Property", property_.property_id, f"[Properties/Property/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

            # check container external id
            if property_.container and not re.match(view_id_compliance_regex, property_.container.external_id):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Container", property_.container.external_id, f"[Properties/Container/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

            # check container property external id
            if property_.container_property and not re.match(
                dms_property_id_compliance_regex, property_.container_property
            ):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Container Property", property_.container_property, f"[Properties/Container Property/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

            # expected value type, as it is case sensitive should be ok
            if not re.match(value_id_compliance_regex, property_.expected_value_type.external_id):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Value type", property_.expected_value_type.external_id, f"[Properties/Type/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

    if return_report:
        return flag, wrangle_warnings(validation_warnings)
    else:
        return flag


@overload
def are_properties_redefined(transformation_rules: Rules, return_report: Literal[True]) -> tuple[bool, list[dict]]: ...


@overload
def are_properties_redefined(transformation_rules: Rules, return_report: Literal[False] = False) -> bool: ...


def are_properties_redefined(
    transformation_rules: Rules, return_report: bool = False
) -> bool | tuple[bool, list[dict]]:
    flag: bool = False
    with warnings.catch_warnings(record=True) as validation_warnings:
        analyzed_properties = {}
        for property_ in transformation_rules.properties.values():
            if property_.property_id not in analyzed_properties:
                analyzed_properties[property_.property_id] = [property_.class_id]
            elif property_.class_id in analyzed_properties[property_.property_id]:
                flag = True
                warnings.warn(
                    exceptions.PropertyRedefined(property_.property_id, property_.class_id).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )

            else:
                analyzed_properties[property_.property_id].append(property_.class_id)

    if return_report:
        return flag, wrangle_warnings(validation_warnings)
    else:
        return flag


def property_ids_camel_case_compliant(transformation_rules) -> bool | tuple[bool, list[dict]]:
    raise NotImplementedError()


def class_id_pascal_case_compliant(transformation_rules) -> bool | tuple[bool, list[dict]]:
    raise NotImplementedError()



import re
from typing import Any, Literal, TypeAlias, cast, overload

from pydantic import HttpUrl, TypeAdapter, ValidationError
from rdflib import Literal as RdfLiteral
from rdflib import Namespace, URIRef

Triple: TypeAlias = tuple[URIRef, URIRef, RdfLiteral | URIRef]


@overload
def remove_namespace_from_uri(
    *URI: URIRef | str,
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> str: ...


@overload
def remove_namespace_from_uri(
    *URI: tuple[URIRef | str, ...],
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> tuple[str, ...]: ...


def remove_namespace_from_uri(
    *URI: URIRef | str | tuple[URIRef | str, ...],
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> tuple[str, ...] | str:
    """Removes namespace from URI

    Args
        URI: URIRef | str
            URI of an entity
        special_separator : str
            Special separator to use instead of # or / if present in URI
            Set by default to "#_" which covers special client use case
        validation: str
            Validation type to use for URI. If set to "full", URI will be validated using pydantic
            If set to "prefix", only check if URI starts with http or https will be made

    Returns
        Entities id without namespace

    Examples:

        >>> remove_namespace_from_uri("http://www.example.org/index.html#section2")
        'section2'
        >>> remove_namespace_from_uri("http://www.example.org/index.html#section2", "http://www.example.org/index.html#section3")
        ('section2', 'section3')
    """
    if isinstance(URI, str | URIRef):
        uris = (URI,)
    elif isinstance(URI, tuple):
        # Assume that all elements in the tuple are of the same type following type hint
        uris = cast(tuple[URIRef | str, ...], URI)
    else:
        raise TypeError(f"URI must be of type URIRef or str, got {type(URI)}")

    output = []
    for u in uris:
        if validation == "full":
            try:
                _ = TypeAdapter(HttpUrl).validate_python(u)
                output.append(u.split(special_separator if special_separator in u else "#" if "#" in u else "/")[-1])
            except ValidationError:
                output.append(str(u))
        else:
            if u.lower().startswith("http"):
                output.append(u.split(special_separator if special_separator in u else "#" if "#" in u else "/")[-1])
            else:
                output.append(str(u))

    return tuple(output) if len(output) > 1 else output[0]


def get_namespace(URI: URIRef, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    str
        Entity id without namespace
    """
    if special_separator in URI:
        return URI.split(special_separator)[0] + special_separator
    elif "#" in URI:
        return URI.split("#")[0] + "#"
    else:
        return "/".join(URI.split("/")[:-1]) + "/"


def as_neat_compliant_uri(uri: URIRef) -> URIRef:
    namespace = get_namespace(uri)
    id_ = remove_namespace_from_uri(uri)
    compliant_uri = re.sub(r"[^a-zA-Z0-9-_.]", "", id_)
    return URIRef(f"{namespace}{compliant_uri}")


def convert_rdflib_content(content: RdfLiteral | URIRef | dict | list) -> Any:
    if isinstance(content, RdfLiteral) or isinstance(content, URIRef):
        return content.toPython()
    elif isinstance(content, dict):
        return {key: convert_rdflib_content(value) for key, value in content.items()}
    elif isinstance(content, list):
        return [convert_rdflib_content(item) for item in content]
    else:
        return content


def uri_to_short_form(URI: URIRef, prefixes: dict[str, Namespace]) -> str | URIRef:
    """Returns the short form of a URI if its namespace is present in the prefixes dict,
    otherwise returns the URI itself

    Args:
        URI: URI to be converted to form prefix:entityName
        prefixes: dict of prefixes

    Returns:
        shortest form of the URI if its namespace is present in the prefixes dict,
        otherwise returns the URI itself
    """
    uris: set[str | URIRef] = {URI}
    for prefix, namespace in prefixes.items():
        if URI.startswith(namespace):
            uris.add(f"{prefix}:{URI.replace(namespace, '')}")
    return min(uris, key=len)


def _traverse(hierarchy: dict, graph: dict, names: list[str]) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def get_inheritance_path(child: Any, child_parent: dict[Any, list[Any]]) -> list:
    """Returns the inheritance path for a given child

    Args:
        child: Child class
        child_parent: Dictionary of child to parent relationship

    Returns:
        Inheritance path for a given child

    !!! note "No Circular Inheritance"
        This method assumes that the child_parent dictionary is a tree and does not contain any cycles.
    """
    path = []
    if child in child_parent:
        path.extend(child_parent[child])
        for parent in child_parent[child]:
            path.extend(get_inheritance_path(parent, child_parent))
    return path



import sys
import warnings
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

import pandas as pd
from rdflib import Graph, Namespace, URIRef
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph._shared import MIMETypes
from cognite.neat.graph.extractors import RdfFileExtractor, TripleExtractors
from cognite.neat.graph.models import Triple
from cognite.neat.graph.queries import Queries
from cognite.neat.graph.transformers import Transformers
from cognite.neat.rules.analysis import InformationAnalysis
from cognite.neat.rules.models import InformationRules
from cognite.neat.rules.models.entities import ClassEntity
from cognite.neat.utils.auxiliary import local_import

from ._provenance import Change, Provenance

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class NeatGraphStore:
    """NeatGraphStore is a class that stores the graph and provides methods to read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        rules:
    """

    rdf_store_type: str

    def __init__(
        self,
        graph: Graph,
        rules: InformationRules | None = None,
    ):
        self.rules: InformationRules | None = None

        _start = datetime.now(timezone.utc)
        self.graph = graph
        self.provenance = Provenance(
            [
                Change.record(
                    activity=f"{type(self).__name__}.__init__",
                    start=_start,
                    end=datetime.now(timezone.utc),
                    description=f"Initialize graph store as {type(self.graph.store).__name__}",
                )
            ]
        )

        if rules:
            self.add_rules(rules)
        else:
            self.base_namespace = DEFAULT_NAMESPACE

        self.queries = Queries(self.graph, self.rules)

    def add_rules(self, rules: InformationRules) -> None:
        """This method is used to add rules to the graph store and it is the only correct
        way to add rules to the graph store, after the graph store has been initialized.
        """

        self.rules = rules
        self.base_namespace = self.rules.metadata.namespace
        self.queries = Queries(self.graph, self.rules)
        self.provenance.append(
            Change.record(
                activity=f"{type(self)}.rules",
                start=datetime.now(timezone.utc),
                end=datetime.now(timezone.utc),
                description=f"Added rules to graph store as {type(self.rules).__name__}",
            )
        )

        if self.rules.prefixes:
            self._upsert_prefixes(self.rules.prefixes)

    def _upsert_prefixes(self, prefixes: dict[str, Namespace]) -> None:
        """Adds prefixes to the graph store."""
        _start = datetime.now(timezone.utc)
        for prefix, namespace in prefixes.items():
            self.graph.bind(prefix, namespace)

        self.provenance.append(
            Change.record(
                activity=f"{type(self).__name__}._upsert_prefixes",
                start=_start,
                end=datetime.now(timezone.utc),
                description="Upsert prefixes to graph store",
            )
        )

    @classmethod
    def from_memory_store(cls, rules: InformationRules | None = None) -> "Self":
        return cls(Graph(), rules)

    @classmethod
    def from_sparql_store(
        cls,
        query_endpoint: str | None = None,
        update_endpoint: str | None = None,
        returnFormat: str = "csv",
        rules: InformationRules | None = None,
    ) -> "Self":
        store = SPARQLUpdateStore(
            query_endpoint=query_endpoint,
            update_endpoint=update_endpoint,
            returnFormat=returnFormat,
            context_aware=False,
            postAsEncoded=False,
            autocommit=False,
        )
        graph = Graph(store=store)
        return cls(graph, rules)

    @classmethod
    def from_oxi_store(cls, storage_dir: Path | None = None, rules: InformationRules | None = None) -> "Self":
        """Creates a NeatGraphStore from an Oxigraph store."""
        local_import("pyoxigraph", "oxi")
        import pyoxigraph

        from cognite.neat.graph.stores._oxrdflib import OxigraphStore

        # Adding support for both oxigraph in-memory and file-based storage
        for i in range(4):
            try:
                oxi_store = pyoxigraph.Store(path=str(storage_dir) if storage_dir else None)
                break
            except OSError as e:
                if "lock" in str(e) and i < 3:
                    continue
                raise e
        else:
            raise Exception("Error initializing Oxigraph store")

        graph = Graph(store=OxigraphStore(store=oxi_store))
        graph.default_union = True

        return cls(graph, rules)

    def write(self, extractor: TripleExtractors) -> None:
        _start = datetime.now(timezone.utc)

        if isinstance(extractor, RdfFileExtractor):
            self._parse_file(extractor.filepath, extractor.mime_type, extractor.base_uri)
        else:
            self._add_triples(extractor.extract())

        self.provenance.append(
            Change.record(
                activity=f"{type(extractor).__name__}",
                start=_start,
                end=datetime.now(timezone.utc),
                description=f"Extracted triples to graph store using {type(extractor).__name__}",
            )
        )

    def read(self, class_: str) -> Iterable[tuple[str, dict[str, list[str]]]]:
        """Read instances for given view from the graph store."""

        if not self.rules:
            warnings.warn("Rules not found in graph store!", stacklevel=2)
            return None

        class_entity = ClassEntity(prefix=self.rules.metadata.prefix, suffix=class_)

        if class_entity not in [definition.class_ for definition in self.rules.classes.data]:
            warnings.warn("Desired type not found in graph!", stacklevel=2)
            return None

        analysis = InformationAnalysis(self.rules)
        has_hop_transformations = analysis.has_hop_transformations()
        has_self_reference_transformations = analysis.has_self_reference_property_transformations()
        if has_hop_transformations or has_self_reference_transformations:
            msg = (
                f"Rules contain [{'Hop' if has_hop_transformations else '' }"
                f", {'SelfReferenceProperty' if has_self_reference_transformations else '' }]"
                " rdfpath."
                f" Run [{'ReduceHopTraversal' if has_hop_transformations else '' }"
                f", {'AddSelfReferenceProperty' if has_self_reference_transformations else '' }]"
                " transformer(s) first!"
            )

            warnings.warn(
                msg,
                stacklevel=2,
            )
            return None

        instance_ids = self.queries.list_instances_ids_of_class(self.rules.metadata.namespace[class_])

        property_renaming_config = InformationAnalysis(self.rules).define_property_renaming_config(class_entity)

        for instance_id in instance_ids:
            yield self.queries.describe(instance_id, property_renaming_config)

    def _parse_file(
        self,
        filepath: Path,
        mime_type: MIMETypes = "application/rdf+xml",
        base_uri: URIRef | None = None,
    ) -> None:
        """Imports graph data from file.

        Args:
            filepath : File path to file containing graph data, by default None
            mime_type : MIME type of graph data, by default "application/rdf+xml"
            add_base_iri : Add base IRI to graph, by default True
        """

        # Oxigraph store, do not want to type hint this as it is an optional dependency
        if type(self.graph.store).__name__ == "OxigraphStore":

            def parse_to_oxi_store():
                local_import("pyoxigraph", "oxi")
                from cognite.neat.graph.stores._oxrdflib import OxigraphStore

                cast(OxigraphStore, self.graph.store)._inner.bulk_load(str(filepath), mime_type, base_iri=base_uri)  # type: ignore[attr-defined]
                cast(OxigraphStore, self.graph.store)._inner.optimize()  # type: ignore[attr-defined]

            parse_to_oxi_store()

        # All other stores
        else:
            if filepath.is_file():
                self.graph.parse(filepath, publicID=base_uri)
            else:
                for filename in filepath.iterdir():
                    if filename.is_file():
                        self.graph.parse(filename, publicID=base_uri)

    def _add_triples(self, triples: Iterable[Triple], batch_size: int = 10_000):
        """Adds triples to the graph store in batches.

        Args:
            triples: list of triples to be added to the graph store
            batch_size: Batch size of triples per commit, by default 10_000
            verbose: Verbose mode, by default False
        """

        commit_counter = 0
        number_of_written_triples = 0

        def check_commit(force_commit: bool = False):
            """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
            nonlocal commit_counter
            nonlocal number_of_written_triples
            if force_commit:
                number_of_written_triples += commit_counter
                self.graph.commit()
                return
            commit_counter += 1
            if commit_counter >= batch_size:
                number_of_written_triples += commit_counter
                self.graph.commit()
                commit_counter = 0

        for triple in triples:
            self.graph.add(triple)
            check_commit()

        check_commit(force_commit=True)

    def transform(self, transformer: Transformers) -> None:
        """Transforms the graph store using a transformer."""

        missing_changes = [
            change for change in transformer._need_changes if not self.provenance.activity_took_place(change)
        ]
        if self.provenance.activity_took_place(type(transformer).__name__) and transformer._use_only_once:
            warnings.warn(
                f"Cannot transform graph store with {type(transformer).__name__}, already applied",
                stacklevel=2,
            )
        elif missing_changes:
            warnings.warn(
                (
                    f"Cannot transform graph store with {type(transformer).__name__}, "
                    f"missing one or more required changes [{', '.join(missing_changes)}]"
                ),
                stacklevel=2,
            )

        else:
            _start = datetime.now(timezone.utc)
            transformer.transform(self.graph)
            self.provenance.append(
                Change.record(
                    activity=f"{type(transformer).__name__}",
                    start=_start,
                    end=datetime.now(timezone.utc),
                    description=transformer.description,
                )
            )

    @property
    def summary(self) -> pd.DataFrame:
        return pd.DataFrame(self.queries.summarize_instances(), columns=["Type", "Occurrence"])

    def _repr_html_(self) -> str:
        provenance = self.provenance._repr_html_()
        summary: pd.DataFrame = self.summary
        summary_text = (
            "<br /><strong>Graph is empty</strong><br />"
            if summary.empty
            else f"<br /><strong>Graph content</strong><br />{cast(pd.DataFrame, summary)._repr_html_()}"  # type: ignore[operator]
        )

        return (
            f"<strong>{type(self).__name__}</strong> A graph store is a container for storing triples. "
            "It can be queried and transformed to extract information.<br />"
            "<strong>Provenance</strong> Provenance is a record of changes that have occurred in the graph store.<br />"
            f"{provenance}"
            f"{summary_text}"
        )



import sys
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

from pydantic import Field, field_validator, model_validator
from pydantic.main import IncEx
from rdflib import Namespace

from cognite.neat.constants import get_default_prefixes
from cognite.neat.issues import MultiValueError
from cognite.neat.rules import issues
from cognite.neat.rules.models._base import BaseRules, RoleTypes, SheetList
from cognite.neat.rules.models.domain import DomainRules
from cognite.neat.rules.models.entities import (
    CdfResourceEntityList,
    ClassEntity,
    MultiValueTypeInfo,
    Undefined,
)
from cognite.neat.rules.models.information import (
    InformationClass,
    InformationMetadata,
    InformationProperty,
    InformationRules,
)

if TYPE_CHECKING:
    from cognite.neat.rules.models.dms._rules import DMSRules


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class AssetMetadata(InformationMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.asset


class AssetClass(InformationClass): ...


class AssetProperty(InformationProperty):
    """
    A property is a characteristic of a class. It is a named attribute of a class that describes a range of values
    or a relationship to another class.

    Args:
        class_: Class ID to which property belongs
        property_: Property ID of the property
        name: Property name.
        value_type: Type of value property will hold (data or link to another class)
        min_count: Minimum count of the property values. Defaults to 0
        max_count: Maximum count of the property values. Defaults to None
        default: Default value of the property
        reference: Reference to the source of the information, HTTP URI
        match_type: The match type of the resource being described and the source entity.
        transformation: Actual rule for the transformation from source to target representation of
              knowledge graph. Defaults to None (no transformation)
        implementation: Details on how given class-property is implemented in the classic CDF
    """

    implementation: CdfResourceEntityList = Field(alias="Implementation")


class AssetRules(BaseRules):
    metadata: AssetMetadata = Field(alias="Metadata")
    properties: SheetList[AssetProperty] = Field(alias="Properties")
    classes: SheetList[AssetClass] = Field(alias="Classes")
    prefixes: dict[str, Namespace] = Field(default_factory=get_default_prefixes)
    last: "AssetRules | None" = Field(None, alias="Last")
    reference: "AssetRules | None" = Field(None, alias="Reference")

    @field_validator("prefixes", mode="before")
    def parse_str(cls, values: Any) -> Any:
        if isinstance(values, dict):
            return {key: Namespace(value) if isinstance(value, str) else value for key, value in values.items()}
        return values

    @model_validator(mode="after")
    def update_entities_prefix(self) -> Self:
        # update expected_value_types
        for property_ in self.properties:
            if isinstance(property_.value_type, ClassEntity) and property_.value_type.prefix is Undefined:
                property_.value_type.prefix = self.metadata.prefix

            if isinstance(property_.value_type, MultiValueTypeInfo):
                property_.value_type.set_default_prefix(self.metadata.prefix)

            if property_.class_.prefix is Undefined:
                property_.class_.prefix = self.metadata.prefix

        # update parent classes
        for class_ in self.classes:
            if class_.parent:
                for parent in class_.parent:
                    if not isinstance(parent.prefix, str):
                        parent.prefix = self.metadata.prefix
            if class_.class_.prefix is Undefined:
                class_.class_.prefix = self.metadata.prefix

        return self

    @model_validator(mode="after")
    def post_validation(self) -> "AssetRules":
        from ._validation import AssetPostValidation

        issue_list = AssetPostValidation(cast(InformationRules, self)).validate()
        if issue_list.warnings:
            issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError([error for error in issue_list if isinstance(error, issues.NeatValidationError)])
        return self

    def dump(
        self,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        as_reference: bool = False,
    ) -> dict[str, Any]:
        from ._serializer import _AssetRulesSerializer

        dumped = self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        prefix = self.metadata.prefix
        serializer = _AssetRulesSerializer(by_alias, prefix)
        cleaned = serializer.clean(dumped, as_reference)
        last = "Last" if by_alias else "last"
        if last_dump := cleaned.get(last):
            cleaned[last] = serializer.clean(last_dump, False)
        reference = "Reference" if by_alias else "reference"
        if self.reference and (ref_dump := cleaned.get(reference)):
            prefix = self.reference.metadata.prefix
            cleaned[reference] = _AssetRulesSerializer(by_alias, prefix).clean(ref_dump, True)
        return cleaned

    def as_domain_rules(self) -> DomainRules:
        from ._converter import _AssetRulesConverter

        return _AssetRulesConverter(self.as_information_rules()).as_domain_rules()

    def as_dms_rules(self) -> "DMSRules":
        from ._converter import _AssetRulesConverter

        return _AssetRulesConverter(self.as_information_rules()).as_dms_rules()

    def as_information_rules(self) -> InformationRules:
        return InformationRules.model_validate(self.model_dump())



import time
from pathlib import Path
from typing import ClassVar, cast

from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import (
    SolutionGraph,
    SourceGraph,
)
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "GraphToRdfFile",
]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class GraphToRdfFile(Step):
    """
    The step generates loads graph to RDF file
    """

    description = "The step generates nodes and edges from the graph"
    category = CATEGORY
    version = "private-beta"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Graph",
            value="source",
            options=["source", "solution"],
            label=("The graph to be used for loading RDF File." " Supported options : source, solution"),
        ),
        Configurable(
            name="File path",
            value="staging/graph_export.ttl",
            label=("Relative path for the RDF file storage, " "must end with .ttl !"),
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, graph: SourceGraph | SolutionGraph
    ) -> FlowMessage:  # type: ignore[syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        storage_path = self.data_store_path / Path(self.configs["File path"])
        relative_graph_file_path = str(storage_path).split("/data/")[1]

        graph_name = self.configs["Graph"] or "source"

        if graph_name == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SourceGraph"])

        graph.graph.serialize(str(storage_path), format="turtle")

        output_text = (
            "<p></p>"
            "Graph loaded to RDF file can be downloaded here : "
            f'<a href="/data/{relative_graph_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)



from pathlib import Path

# we should make this a proper package that loads examples
# similar how they handle it in xarray:
# https://github.com/pydata/xarray/blob/main/xarray/tutorial.py
# Currently there are simple paths to the examples which are then easily loaded in the notebooks
nordic44_knowledge_graph = Path(__file__).parent / "Knowledge-Graph-Nordic44.xml"
nordic44_knowledge_graph_dirty = (
    Path(__file__).parent / "Knowledge-Graph-Nordic44-dirty.xml"
)



from collections.abc import Iterable
from typing import cast

from cognite.client import CogniteClient
from prometheus_client import REGISTRY, Counter, Gauge, Metric


class NeatMetricsCollector:
    def __init__(self, name: str, cdf_client: CogniteClient | None = None) -> None:
        self.name = name
        self.metrics: dict[str, Gauge | Counter] = {}

    def register_metric(
        self,
        metric_name: str,
        metric_description: str = "",
        m_type: str = "gauge",
        metric_labels: list[str] | None = None,
    ) -> Gauge | Counter | None:
        """Register metric in prometheus"""
        metric_name = self.sanitize_metric_name(metric_name)
        metric_labels = [] if metric_labels is None else metric_labels

        metric_name = f"neat_workflow_{self.sanitize_metric_name(self.name)}_{metric_name}"
        if metric_name in REGISTRY._names_to_collectors:
            self.metrics[metric_name] = cast(Gauge | Counter, REGISTRY._names_to_collectors[metric_name])
            return self.metrics[metric_name]

        metric: Gauge | Counter | None = None
        if m_type == "gauge":
            metric = Gauge(metric_name, metric_description, metric_labels)
        elif m_type == "counter":
            metric = Counter(metric_name, metric_description, metric_labels)

        if metric:
            self.metrics[metric_name] = metric
            return metric
        return None

    def get(self, metric_name: str, labels: dict[str, str] | None = None) -> Gauge | Counter | None:
        """Return metric by name"""
        metric_name = self.sanitize_metric_name(metric_name)
        labels = {} if labels is None else labels
        metric_name = f"neat_workflow_{self.sanitize_metric_name(self.name)}_{metric_name}"
        if metric_name in self.metrics:
            return self.metrics[metric_name].labels(**(labels or {}))
        return None

    def report_metric_value(
        self,
        metric_name: str,
        metric_description: str = "",
        m_type: str = "gauge",
        labels: dict[str, str] | None = None,
    ) -> Gauge | Counter | None:
        self.sanitize_metric_name(metric_name)
        metric = self.register_metric(metric_name, metric_description, m_type, [k for k, v in (labels or {}).items()])
        if metric:
            return metric.labels(**(labels or {}))
        return None

    def sanitize_metric_name(self, metric_name: str) -> str:
        return metric_name.replace("-", "_").replace(" ", "_").replace(".", "_").replace(":", "_").lower()

    def collect(self) -> Iterable[Metric]:
        return cast(Iterable[Metric], self.metrics.values())

    def report_to_cdf(self):
        pass



import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import total_ordering
from typing import Any, ClassVar

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling import ContainerId, ViewId
from pydantic_core import ErrorDetails
from rdflib import Namespace

from cognite.neat.issues import MultiValueError
from cognite.neat.utils.spreadsheet import SpreadsheetRead

from .base import DefaultPydanticError, NeatValidationError

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = [
    "InvalidSheetError",
    "InvalidRowError",
    "InvalidPropertyError",
    "InvalidClassError",
    "PrefixNamespaceCollisionError",
    "InvalidContainerError",
    "InvalidViewError",
    "InvalidRowUnknownSheetError",
    "NonExistingContainerError",
    "NonExistingViewError",
    "ClassNoPropertiesNoParentError",
    "InconsistentContainerDefinitionError",
    "MultiValueTypeError",
    "MultiValueIsListError",
    "MultiNullableError",
    "MultiDefaultError",
    "MultiIndexError",
    "MultiUniqueConstraintError",
]


@dataclass(frozen=True)
class InvalidSheetError(NeatValidationError, ABC):
    @classmethod
    @abstractmethod
    def from_pydantic_error(
        cls,
        error: ErrorDetails,
        read_info_by_sheet: dict[str, SpreadsheetRead] | None = None,
    ) -> Self:
        raise NotImplementedError

    @classmethod
    def from_pydantic_errors(
        cls,
        errors: list[ErrorDetails],
        read_info_by_sheet: dict[str, SpreadsheetRead] | None = None,
        **kwargs: Any,
    ) -> "list[NeatValidationError]":
        output: list[NeatValidationError] = []
        for error in errors:
            if raised_error := error.get("ctx", {}).get("error"):
                if isinstance(raised_error, MultiValueError):
                    for caught_error in raised_error.errors:
                        reader = (read_info_by_sheet or {}).get("Properties", SpreadsheetRead())
                        if isinstance(caught_error, InconsistentContainerDefinitionError):
                            row_numbers = list(caught_error.row_numbers)
                            # The Error classes are immutable, so we have to reuse the set.
                            caught_error.row_numbers.clear()
                            for row_no in row_numbers:
                                # Adjusting the row number to the actual row number in the spreadsheet
                                caught_error.row_numbers.add(reader.adjusted_row_number(row_no))
                        if isinstance(caught_error, InvalidRowError):
                            # Adjusting the row number to the actual row number in the spreadsheet
                            new_row = reader.adjusted_row_number(caught_error.row)
                            # The error is frozen, so we have to use __setattr__ to change the row number
                            object.__setattr__(caught_error, "row", new_row)
                        output.append(caught_error)  # type: ignore[arg-type]
                    continue

            if len(error["loc"]) >= 4:
                sheet_name, *_ = error["loc"]
                error_cls = _INVALID_ROW_ERROR_BY_SHEET_NAME.get(str(sheet_name), InvalidRowUnknownSheetError)
                output.append(error_cls.from_pydantic_error(error, read_info_by_sheet))
                continue

            output.append(DefaultPydanticError.from_pydantic_error(error))
        return output


@dataclass(frozen=True)
@total_ordering
class InvalidRowError(InvalidSheetError, ABC):
    description: ClassVar[str] = "This is a generic class for all invalid row specifications."
    fix: ClassVar[str] = "Follow the instruction in the error message."
    sheet_name: ClassVar[str]

    column: str
    row: int
    type: str
    msg: str
    input: Any
    url: str | None

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, InvalidRowError):
            return NotImplemented
        return (self.sheet_name, self.row, self.column) < (
            other.sheet_name,
            other.row,
            other.column,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InvalidRowError):
            return NotImplemented
        return (self.sheet_name, self.row, self.column) == (
            other.sheet_name,
            other.row,
            other.column,
        )

    @classmethod
    def from_pydantic_error(
        cls,
        error: ErrorDetails,
        read_info_by_sheet: dict[str, SpreadsheetRead] | None = None,
    ) -> Self:
        sheet_name, _, row, column, *__ = error["loc"]
        reader = (read_info_by_sheet or {}).get(str(sheet_name), SpreadsheetRead())
        return cls(
            column=str(column),
            row=reader.adjusted_row_number(int(row)),
            type=error["type"],
            msg=error["msg"],
            input=error.get("input"),
            url=str(url) if (url := error.get("url")) else None,
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["sheet_name"] = self.sheet_name
        output["column"] = self.column
        output["row"] = self.row
        output["type"] = self.type
        output["msg"] = self.msg
        output["input"] = self.input
        output["url"] = self.url
        return output

    def message(self) -> str:
        input_str = str(self.input) if self.input is not None else ""
        input_str = input_str[:50] + "..." if len(input_str) > 50 else input_str
        output = (
            f"In {self.sheet_name}, row={self.row}, column={self.column}: {self.msg}. "
            f"[type={self.type}, input_value={input_str}]"
        )
        if self.url:
            output += f" For further information visit {self.url}"
        return output


@dataclass(frozen=True)
class InvalidPropertyError(InvalidRowError):
    sheet_name = "Properties"


@dataclass(frozen=True)
class InvalidClassError(InvalidRowError):
    sheet_name = "Classes"


@dataclass(frozen=True)
class InvalidContainerError(InvalidRowError):
    sheet_name = "Containers"


@dataclass(frozen=True)
class InvalidViewError(InvalidRowError):
    sheet_name = "Views"


@dataclass(frozen=True)
class InvalidRowUnknownSheetError(InvalidRowError):
    sheet_name = "Unknown"

    actual_sheet_name: str

    @classmethod
    def from_pydantic_error(
        cls,
        error: ErrorDetails,
        read_info_by_sheet: dict[str, SpreadsheetRead] | None = None,
    ) -> Self:
        sheet_name, _, row, column, *__ = error["loc"]
        reader = (read_info_by_sheet or {}).get(str(sheet_name), SpreadsheetRead())
        try:
            return cls(
                column=str(column),
                row=reader.adjusted_row_number(int(row)),
                actual_sheet_name=str(sheet_name),
                type=error["type"],
                msg=error["msg"],
                input=error.get("input"),
                url=str(url) if (url := error.get("url")) else None,
            )
        except ValueError:
            return DefaultPydanticError.from_pydantic_error(error)  # type: ignore[return-value]

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["actual_sheet_name"] = self.actual_sheet_name
        return output


_INVALID_ROW_ERROR_BY_SHEET_NAME = {
    cls_.sheet_name: cls_ for cls_ in InvalidRowError.__subclasses__() if cls_ is not InvalidRowError
}


@dataclass(frozen=True)
class NonExistingContainerError(InvalidPropertyError):
    description = "The container referenced by the property is missing in the container sheet"
    fix = "Add the container to the container sheet"

    container_id: ContainerId

    def message(self) -> str:
        return (
            f"In {self.sheet_name}, row={self.row}, column={self.column}: The container with "
            f"id {self.container_id} is missing in the container sheet."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container_id"] = self.container_id
        return output


@dataclass(frozen=True)
class NonExistingViewError(InvalidPropertyError):
    description = "The view referenced by the property is missing in the view sheet"
    fix = "Add the view to the view sheet"

    view_id: ViewId

    def message(self) -> str:
        return (
            f"In {self.sheet_name}, row={self.row}, column={self.column}: The view with "
            f"id {self.view_id} is missing in the view sheet."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id
        return output


@dataclass(frozen=True)
class PropertiesDefinedForUndefinedClassesError(NeatValidationError):
    description = "Properties are defined for undefined classes."
    fix = "Make sure to define class in the Classes sheet."

    classes: list[str]

    def dump(self) -> dict[str, list[str]]:
        output = super().dump()
        output["classes"] = self.classes
        return output

    def message(self) -> str:
        return (
            f"Classes {', '.join(self.classes)} have properties assigned to them, but"
            " they are not defined in the Classes sheet."
        )


@dataclass(frozen=True)
class ClassNoPropertiesNoParentError(NeatValidationError):
    description = "Class has no properties and no parents."
    fix = "Check if the class should have properties or parents."

    classes: list[str]

    def dump(self) -> dict[str, list[str]]:
        output = super().dump()
        output["classes"] = self.classes
        return output

    def message(self) -> str:
        if len(self.classes) > 1:
            return f"Classes {', '.join(self.classes)} have no direct or inherited properties. This may be a mistake."
        return f"Class {self.classes[0]} have no direct or inherited properties. This may be a mistake."


@dataclass(frozen=True)
class AssetRulesHaveCircularDependencyError(NeatValidationError):
    description = "Asset rules have circular dependencies."
    fix = "Linking between classes via property that maps to parent_external_id must yield hierarchy structure."

    classes: list[str]

    def dump(self) -> dict[str, list[tuple[str, str]]]:
        output = super().dump()
        output["classes"] = self.classes
        return output

    def message(self) -> str:
        return f"Asset rules have circular dependencies between classes {', '.join(self.classes)}."


@dataclass(frozen=True)
class AssetParentPropertyPointsToDataValueTypeError(NeatValidationError):
    description = "Parent property points to a data value type instead of a class."
    fix = "Make sure that the parent property points to a class."

    class_property_with_data_value_type: list[tuple[str, str]]

    def dump(self) -> dict[str, list[tuple[str, str]]]:
        output = super().dump()
        output["class_property"] = self.class_property_with_data_value_type
        return output

    def message(self) -> str:
        text = [
            f"class {class_} property {property_}" for class_, property_ in self.class_property_with_data_value_type
        ]
        return f"Following  {', and'.join(text)} point to data value type instead to classes. This is a mistake."


@dataclass(frozen=True)
class ParentClassesNotDefinedError(NeatValidationError):
    description = "Parent classes are not defined."
    fix = "Check if the parent classes are defined in Classes sheet."

    classes: list[str]

    def dump(self) -> dict[str, list[str]]:
        output = super().dump()
        output["classes"] = self.classes
        return output

    def message(self) -> str:
        if len(self.classes) > 1:
            return f"Parent classes {', '.join(self.classes)} are not defined. This may be a mistake."
        return f"Parent classes {', '.join(self.classes[0])} are not defined. This may be a mistake."


@dataclass(frozen=True)
class PrefixNamespaceCollisionError(NeatValidationError):
    description = "Same namespaces are assigned to different prefixes."
    fix = "Make sure that each unique namespace is assigned to a unique prefix"

    namespaces: list[Namespace]
    prefixes: list[str]

    def dump(self) -> dict[str, list[str]]:
        output = super().dump()
        output["prefixes"] = self.prefixes
        output["namespaces"] = self.namespaces
        return output

    def message(self) -> str:
        return (
            f"Namespaces {', '.join(self.namespaces)} are assigned multiple times."
            f" Impacted prefixes: {', '.join(self.prefixes)}."
        )


@dataclass(frozen=True)
class ValueTypeNotDefinedError(NeatValidationError):
    description = "Value types referred by properties are not defined in Rules."
    fix = "Make sure that all value types are defined in Rules."

    value_types: list[str]

    def dump(self) -> dict[str, list[str]]:
        output = super().dump()
        output["classes"] = self.value_types
        return output

    def message(self) -> str:
        if len(self.value_types) > 1:
            return f"Value types {', '.join(self.value_types)} are not defined. This may be a mistake."
        return f"Value types {', '.join(self.value_types[0])} are not defined. This may be a mistake."


@dataclass(frozen=True)
class InconsistentContainerDefinitionError(NeatValidationError, ABC):
    description = "This is a base class for all errors related to inconsistent container definitions"
    fix = "Ensure all properties using the same container have the same type, constraints, and indexes."
    container: dm.ContainerId
    property_name: str
    row_numbers: set[int]

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output.update(
            {
                "container": self.container.dump(),
                "property_name": self.property_name,
                "row_numbers": sorted(self.row_numbers),
            }
        )
        return output


@dataclass(frozen=True)
class MultiValueTypeError(InconsistentContainerDefinitionError):
    description = "The property has multiple value types"
    fix = "Use the same value type for all properties using the same container."
    value_types: set[str]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different value types: {self.value_types}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["value_types"] = sorted(self.value_types)
        return output


@dataclass(frozen=True)
class MultiValueIsListError(InconsistentContainerDefinitionError):
    description = "The property has multiple list definitions"
    fix = "Use the same list definition for all properties using the same container."
    list_definitions: set[bool]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different list definitions: {self.list_definitions}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["list_definitions"] = sorted(self.list_definitions)
        return output


@dataclass(frozen=True)
class MultiNullableError(InconsistentContainerDefinitionError):
    description = "The property has multiple nullable definitions"
    fix = "Use the same nullable definition for all properties using the same container."
    nullable_definitions: set[bool]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different nullable definitions: {self.nullable_definitions}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["nullable_definitions"] = sorted(self.nullable_definitions)
        return output


@dataclass(frozen=True)
class MultiDefaultError(InconsistentContainerDefinitionError):
    description = "The property has multiple default definitions"
    fix = "Use the same default definition for all properties using the same container."
    default_definitions: list[str | int | dict | None]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different default definitions: {self.default_definitions}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["default_definitions"] = self.default_definitions
        return output


@dataclass(frozen=True)
class MultiIndexError(InconsistentContainerDefinitionError):
    description = "The property has multiple index definitions"
    fix = "Use the same index definition for all properties using the same container."
    index_definitions: set[str]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different index definitions: {self.index_definitions}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["index_definitions"] = sorted(self.index_definitions)
        return output


@dataclass(frozen=True)
class MultiUniqueConstraintError(InconsistentContainerDefinitionError):
    description = "The property has multiple unique constraint definitions"
    fix = "Use the same unique constraint definition for all properties using the same container."
    unique_constraint_definitions: set[str]

    def message(self) -> str:
        return (
            f"{self.container}.{self.property_name} defined in rows: {sorted(self.row_numbers)} "
            f"has different unique constraint definitions: {self.unique_constraint_definitions}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["unique_constraint_definitions"] = sorted(self.unique_constraint_definitions)
        return output



# we will use prov-o to represent the provenance of the neat graph store
# basically tracking changes that occur in the graph store
# prov-o use concepts of Agent, Activity and Entity to represent provenance
# where in case of neat we have:
# Agent: triples extractors, graph enhancers, contextualizers, etc.
# Activity: write triple, add connection, etc.
# Entity: neat graph store


import uuid
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar

from rdflib import PROV, RDF, Literal, URIRef

from cognite.neat._shared import FrozenNeatObject, NeatList
from cognite.neat.constants import DEFAULT_NAMESPACE


@dataclass(frozen=True)
class Agent:
    id_: URIRef = DEFAULT_NAMESPACE.agent
    acted_on_behalf_of: str = "NEAT"

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.actedOnBehalfOf, self.acted_on_behalf_of),
        ]


@dataclass(frozen=True)
class Activity:
    was_associated_with: Agent
    ended_at_time: datetime
    started_at_time: datetime
    used: str  # this would be set to for example Extractor, Enhancer, Contextualizer, etc.
    id_: URIRef = DEFAULT_NAMESPACE[f"activity-{uuid.uuid4()}"]

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasAssociatedWith, self.was_associated_with.id_),
            (self.id_, PROV.startedAtTime, Literal(self.started_at_time)),
            (self.id_, PROV.endedAtTime, Literal(self.ended_at_time)),
            (self.id_, PROV.used, self.used),
        ]


@dataclass(frozen=True)
class Entity:
    was_generated_by: Activity
    was_attributed_to: Agent
    id_: URIRef = DEFAULT_NAMESPACE["graph-store"]

    def as_triples(self):
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasGeneratedBy, self.was_generated_by.id_),
            (self.id_, PROV.wasAttributedTo, self.was_attributed_to.id_),
        ]


@dataclass(frozen=True)
class Change(FrozenNeatObject):
    agent: Agent
    activity: Activity
    entity: Entity
    description: str

    def as_triples(self):
        return self.agent.as_triples() + self.activity.as_triples() + self.entity.as_triples()

    @classmethod
    def record(cls, activity: str, start: datetime, end: datetime, description: str):
        """User friendly method to record a change that occurred in the graph store."""
        agent = Agent()
        activity = Activity(used=activity, was_associated_with=agent, started_at_time=start, ended_at_time=end)
        entity = Entity(was_generated_by=activity, was_attributed_to=agent)
        return cls(agent, activity, entity, description)

    def dump(self, aggregate: bool = True) -> dict[str, str]:
        return {
            "Agent": self.agent.id_,
            "Activity": self.activity.id_,
            "Entity": self.entity.id_,
            "Description": self.description,
        }


T_Change = TypeVar("T_Change", bound=Change)


class Provenance(NeatList[Change]):
    def __init__(self, changes: Sequence[T_Change] | None = None):
        super().__init__(changes or [])

    def activity_took_place(self, activity: str) -> bool:
        return any(change.activity.used == activity for change in self)

    def __delitem__(self, *args, **kwargs):
        raise TypeError("Cannot delete change from provenance")

    def __setitem__(self, *args, **kwargs):
        raise TypeError("Cannot modify change from provenance")



import warnings
from graphlib import TopologicalSorter
from typing import cast

from cognite.neat.rules.models import AssetRules
from cognite.neat.rules.models._rdfpath import RDFPath
from cognite.neat.rules.models.asset import AssetClass, AssetProperty
from cognite.neat.rules.models.entities import (
    AssetEntity,
    AssetFields,
    ClassEntity,
    EntityTypes,
    ReferenceEntity,
    RelationshipEntity,
)

from ._base import BaseAnalysis


class AssetAnalysis(BaseAnalysis[AssetRules, AssetClass, AssetProperty, ClassEntity, str]):
    """Assumes analysis over only the complete schema"""

    def _get_reference(self, class_or_property: AssetClass | AssetProperty) -> ReferenceEntity | None:
        return class_or_property.reference if isinstance(class_or_property.reference, ReferenceEntity) else None

    def _get_cls_entity(self, class_: AssetClass | AssetProperty) -> ClassEntity:
        return class_.class_

    def _get_prop_entity(self, property_: AssetProperty) -> str:
        return property_.property_

    def _get_cls_parents(self, class_: AssetClass) -> list[ClassEntity] | None:
        return list(class_.parent or []) or None

    def _get_reference_rules(self) -> AssetRules | None:
        return self.rules.reference

    @classmethod
    def _set_cls_entity(cls, property_: AssetProperty, class_: ClassEntity) -> None:
        property_.class_ = class_

    def _get_object(self, property_: AssetProperty) -> ClassEntity | None:
        return property_.value_type if isinstance(property_.value_type, ClassEntity) else None

    def _get_max_occurrence(self, property_: AssetProperty) -> int | float | None:
        return property_.max_count

    def _get_classes(self) -> list[AssetClass]:
        return list(self.rules.classes)

    def _get_properties(self) -> list[AssetProperty]:
        return list(self.rules.properties)

    def subset_rules(self, desired_classes: set[ClassEntity]) -> AssetRules:
        raise NotImplementedError("Method not implemented")

    def class_property_pairs(
        self,
        only_rdfpath: bool = False,
        consider_inheritance: bool = False,
        implementation_type: EntityTypes = EntityTypes.asset,
    ) -> dict[ClassEntity, dict[str, AssetProperty]]:
        class_property_pairs = {}

        T_implementation = AssetEntity if implementation_type == EntityTypes.asset else RelationshipEntity

        for class_, properties in self.classes_with_properties(consider_inheritance).items():
            processed_properties = {}
            for property_ in properties:
                if property_.property_ in processed_properties:
                    # TODO: use appropriate Warning class from _exceptions.py
                    # if missing make one !
                    warnings.warn(
                        f"Property {property_.property_} for {class_} has been defined more than once!"
                        " Only the first definition will be considered, skipping the rest..",
                        stacklevel=2,
                    )
                    continue

                if (
                    property_.implementation
                    and any(isinstance(implementation, T_implementation) for implementation in property_.implementation)
                    and (not only_rdfpath or (only_rdfpath and isinstance(property_.transformation, RDFPath)))
                ):
                    implementation = [
                        implementation
                        for implementation in property_.implementation
                        if isinstance(implementation, T_implementation)
                    ]

                    processed_properties[property_.property_] = property_.model_copy(
                        deep=True, update={"implementation": implementation}
                    )

            if processed_properties:
                class_property_pairs[class_] = processed_properties

        return class_property_pairs

    def class_topological_sort(self) -> list[ClassEntity]:
        child_parent_asset: dict[ClassEntity, set[ClassEntity]] = {}
        for class_, properties in self.asset_definition().items():
            child_parent_asset[class_] = set()
            for property_ in properties.values():
                if any(
                    cast(AssetEntity, implementation).property_ == AssetFields.parentExternalId
                    for implementation in property_.implementation
                ):
                    child_parent_asset[property_.class_].add(cast(ClassEntity, property_.value_type))

        return list(TopologicalSorter(child_parent_asset).static_order())

    def asset_definition(
        self, only_rdfpath: bool = False, consider_inheritance: bool = False
    ) -> dict[ClassEntity, dict[str, AssetProperty]]:
        return self.class_property_pairs(
            consider_inheritance=consider_inheritance,
            only_rdfpath=only_rdfpath,
            implementation_type=EntityTypes.asset,
        )

    def relationship_definition(
        self, only_rdfpath: bool = False, consider_inheritance: bool = False
    ) -> dict[ClassEntity, dict[str, AssetProperty]]:
        return self.class_property_pairs(
            consider_inheritance=consider_inheritance,
            only_rdfpath=only_rdfpath,
            implementation_type=EntityTypes.relationship,
        )

    def define_property_renaming_config(self, class_: ClassEntity) -> dict[str, str]:
        property_renaming_configuration = {}

        if asset_definition := self.asset_definition().get(class_, None):
            for property_, transformation in asset_definition.items():
                asset_property = cast(list[AssetEntity], transformation.implementation)[0].property_

                if asset_property != "metadata":
                    property_renaming_configuration[property_] = str(asset_property)
                else:
                    property_renaming_configuration[property_] = f"{asset_property}.{property_}"

        return property_renaming_configuration



from abc import ABC
from dataclasses import dataclass, field
from functools import total_ordering
from typing import Any, Generic

from cognite.neat._shared import T_ID, NeatList, NeatObject
from cognite.neat.issues import NeatIssueList


@total_ordering
@dataclass
class UploadResultCore(NeatObject, ABC):
    name: str
    error_messages: list[str] = field(default_factory=list)
    issues: NeatIssueList = field(default_factory=NeatIssueList)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, UploadResultCore):
            return self.name < other.name
        else:
            return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, UploadResultCore):
            return self.name == other.name
        else:
            return NotImplemented

    def dump(self, aggregate: bool = True) -> dict[str, Any]:
        return {"name": self.name}


class UploadResultList(NeatList[UploadResultCore]): ...


@dataclass
class UploadResult(UploadResultCore, Generic[T_ID]):
    created: set[T_ID] = field(default_factory=set)
    upserted: set[T_ID] = field(default_factory=set)
    deleted: set[T_ID] = field(default_factory=set)
    changed: set[T_ID] = field(default_factory=set)
    unchanged: set[T_ID] = field(default_factory=set)
    skipped: set[T_ID] = field(default_factory=set)
    failed_created: set[T_ID] = field(default_factory=set)
    failed_upserted: set[T_ID] = field(default_factory=set)
    failed_changed: set[T_ID] = field(default_factory=set)
    failed_deleted: set[T_ID] = field(default_factory=set)

    @property
    def failed(self) -> int:
        return (
            len(self.failed_created) + len(self.failed_changed) + len(self.failed_deleted) + len(self.failed_upserted)
        )

    @property
    def success(self) -> int:
        return (
            len(self.created)
            + len(self.deleted)
            + len(self.changed)
            + len(self.upserted)
            + len(self.unchanged)
            + len(self.skipped)
        )

    def dump(self, aggregate: bool = True) -> dict[str, Any]:
        output = super().dump(aggregate)
        if self.created:
            output["created"] = len(self.created) if aggregate else list(self.created)
        if self.upserted:
            output["upserted"] = len(self.upserted) if aggregate else list(self.upserted)
        if self.deleted:
            output["deleted"] = len(self.deleted) if aggregate else list(self.deleted)
        if self.changed:
            output["changed"] = len(self.changed) if aggregate else list(self.changed)
        if self.unchanged:
            output["unchanged"] = len(self.unchanged) if aggregate else list(self.unchanged)
        if self.skipped:
            output["skipped"] = len(self.skipped) if aggregate else list(self.skipped)
        if self.failed_created:
            output["failed_created"] = len(self.failed_created) if aggregate else list(self.failed_created)
        if self.failed_upserted:
            output["failed_upserted"] = len(self.failed_upserted) if aggregate else list(self.failed_upserted)
        if self.failed_changed:
            output["failed_changed"] = len(self.failed_changed) if aggregate else list(self.failed_changed)
        if self.failed_deleted:
            output["failed_deleted"] = len(self.failed_deleted) if aggregate else list(self.failed_deleted)
        if self.error_messages:
            output["error_messages"] = len(self.error_messages) if aggregate else self.error_messages
        if self.issues:
            output["issues"] = len(self.issues) if aggregate else [issue.dump() for issue in self.issues]
        return output

    def __str__(self) -> str:
        dumped = self.dump(aggregate=True)
        lines: list[str] = []
        for key, value in dumped.items():
            if key in ["name", "error_messages", "issues"]:
                continue
            lines.append(f"{key}: {value}")
        return f"{self.name.title()}: {', '.join(lines)}"



import logging
import os
import shutil
from pathlib import Path

from rdflib import Graph, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE, get_default_prefixes
from cognite.neat.utils.auxiliary import local_import

from ._base import MIMETypes, NeatGraphStoreBase


class OxiGraphStore(NeatGraphStoreBase):
    """OxiGraph is a class that stores the graph using OxiGraph and provides methods to read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        base_prefix : Used as a base prefix for graph namespace, allowing querying graph data using a shortform of a URI
        namespace : Namespace (aka URI) used to resolve any relative URI in the graph
        prefixes : Dictionary of additional prefixes used and bounded to the graph
    """

    rdf_store_type = "oxigraph"

    def __init__(
        self,
        graph: Graph | None = None,
        base_prefix: str = "",  # usually empty
        namespace: Namespace = DEFAULT_NAMESPACE,
        prefixes: dict[str, Namespace] | None = None,
    ):
        prefixes = prefixes if prefixes else get_default_prefixes()
        super().__init__(graph, base_prefix, namespace, prefixes)

    def _set_graph(self) -> None:
        logging.info("Initializing Oxigraph store")
        local_import("pyoxigraph", "oxi")
        import pyoxigraph

        from cognite.neat.graph.stores import _oxrdflib

        # Adding support for both in-memory and file-based storage
        for i in range(4):
            try:
                oxstore = pyoxigraph.Store(
                    path=(str(self.internal_storage_dir) if self.internal_storage_dir else None)
                )  # Store (Rust object) accepts only str as path and not Path.
                break
            except OSError as e:
                if "lock" in str(e) and i < 3:
                    # lock originated from another instance of the store
                    logging.error("Error initializing Oxigraph store: %s", e)
                else:
                    raise e
        else:
            raise Exception("Error initializing Oxigraph store")

        self.graph = Graph(store=_oxrdflib.OxigraphStore(store=oxstore))
        self.graph.default_union = True
        self.garbage_collector()

    def close(self):
        """Closes the graph."""
        if self.graph is not None:
            try:
                self.graph.store._inner.flush()  # type: ignore[attr-defined]
                self.graph.close(True)
            except Exception as e:
                logging.debug("Error closing graph: %s", e)

    def restart(self):
        """Restarts the graph"""
        self.close()
        self.reinitialize_graph()
        logging.info("GraphStore restarted")

    def import_from_file(
        self,
        graph_file: Path,
        mime_type: MIMETypes = "application/rdf+xml",
        add_base_iri: bool = True,
    ) -> None:
        """Imports graph data from file.

        Args:
            graph_file : File path to file containing graph data, by default None
            mime_type : MIME type of the file, by default "application/rdf+xml"
            add_base_iri : Add base IRI to the graph, by default True
        """
        if add_base_iri:
            self.graph.store._inner.bulk_load(  # type: ignore[attr-defined]
                str(graph_file), mime_type, base_iri=self.namespace
            )
        else:
            self.graph.store._inner.bulk_load(str(graph_file), mime_type)  # type: ignore[attr-defined]
        self.graph.store._inner.optimize()  # type: ignore[attr-defined]
        return None

    def drop(self):
        try:
            self.close()
            # Due to the specifics of Oxigraph, storage directory cannot be deleted immediately
            # after closing the graph and creating a new one
            if self.internal_storage_dir.exists():
                self.storage_dirs_to_delete.append(self.internal_storage_dir)
            self.garbage_collector()

        except Exception as e:
            logging.error(f"Error dropping graph : {e}")

    def garbage_collector(self):
        """Garbage collection of the graph store."""
        # delete all directories in self.storage_dirs_to_delete
        for d in self.storage_dirs_to_delete:
            shutil.rmtree(d)
        self.storage_dirs_to_delete = []

    def __del__(self):
        if self.graph is not None:
            if self.graph.store is not None:
                try:
                    self.graph.store._inner.flush()
                except Exception:
                    logging.debug("Error flushing graph")
            self.graph.close()
        # It requires more investigation os.remove(self.internal_storage_dir / "LOCK")

    def commit(self):
        """Commits the graph."""
        if self.graph:
            if self.graph.store:
                logging.info("Committing graph - flushing and optimizing")
                self.graph.store._inner.flush()
                self.graph.store._inner.optimize()
            self.graph.commit()

    @staticmethod
    def drop_graph_store_storage(storage_path: Path | None) -> None:
        """Drop graph store storage on disk.

        Args:
            storage_path : Path to storage directory
        """
        if storage_path and storage_path.exists():
            for f in os.listdir(storage_path):
                (storage_path / f).unlink()
            logging.info("Graph store dropped.")
        else:
            logging.info(f"Storage path {storage_path} does not exist. Skipping drop.")



import re
import sys
from typing import ClassVar

from cognite.client.data_classes.data_modeling import ViewId
from pydantic import BaseModel, ConfigDict, Field
from rdflib import Literal, URIRef

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self


class EntityTypes(StrEnum):
    subject = "subject"
    predicate = "predicate"
    object = "object"
    class_ = "class"
    property_ = "property"
    object_property = "ObjectProperty"
    data_property = "DatatypeProperty"
    annotation_property = "AnnotationProperty"
    data_value_type = "data_value_type"
    object_value_type = "object_value_type"
    view = "view"
    container = "container"
    undefined = "undefined"


# ALLOWED
ALLOWED_PATTERN = r"[^a-zA-Z0-9-_.]"

# REGEX expressions
PREFIX_REGEX = r"[a-zA-Z]+[a-zA-Z0-9-_.]*[a-zA-Z0-9]+"
SUFFIX_REGEX = r"[a-zA-Z0-9-_.]+[a-zA-Z0-9]|[-_.]*[a-zA-Z0-9]+"
VERSION_REGEX = r"[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])"

ENTITY_ID_REGEX = rf"{PREFIX_REGEX}:({SUFFIX_REGEX})"
ENTITY_ID_REGEX_COMPILED = re.compile(rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})$")
VERSIONED_ENTITY_REGEX_COMPILED = re.compile(
    rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})\(version=(?P<version>{VERSION_REGEX})\)$"
)

CLASS_ID_REGEX = rf"(?P<{EntityTypes.class_}>{ENTITY_ID_REGEX})"
CLASS_ID_REGEX_COMPILED = re.compile(rf"^{CLASS_ID_REGEX}$")

PROPERTY_ID_REGEX = rf"\((?P<{EntityTypes.property_}>{ENTITY_ID_REGEX})\)"
VERSION_ID_REGEX = rf"\(version=(?P<version>{VERSION_REGEX})\)"


class Entity(BaseModel):
    """Entity is a class or property in OWL/RDF sense."""

    model_config: ClassVar[ConfigDict] = ConfigDict(use_enum_values=True)
    prefix: str
    suffix: str
    type_: EntityTypes = Field(default=EntityTypes.undefined)
    name: str | None = None
    description: str | None = None
    version: str | None = None

    @property
    def id(self) -> str:
        return f"{self.prefix}:{self.suffix}"

    @property
    def versioned_id(self) -> str:
        if self.version:
            return f"{self.prefix}:{self.suffix}(version={self.version})"
        else:
            return self.id

    @property
    def space(self) -> str:
        """Returns entity space in CDF."""
        return self.prefix

    @property
    def external_id(self) -> str:
        """Returns entity external id in CDF."""
        return self.suffix

    def __repr__(self):
        return self.id

    @classmethod
    def from_string(cls, entity_string: str, base_prefix: str | None = None, **kwargs) -> Self:
        if result := VERSIONED_ENTITY_REGEX_COMPILED.match(entity_string):
            return cls(
                prefix=result.group("prefix"),
                suffix=result.group("suffix"),
                name=result.group("suffix"),
                version=result.group("version"),
                **kwargs,
            )
        elif result := ENTITY_ID_REGEX_COMPILED.match(entity_string):
            return cls(
                prefix=result.group("prefix"), suffix=result.group("suffix"), name=result.group("suffix"), **kwargs
            )
        elif base_prefix and re.match(SUFFIX_REGEX, entity_string) and re.match(PREFIX_REGEX, base_prefix):
            return cls(prefix=base_prefix, suffix=entity_string, name=entity_string, **kwargs)
        else:
            raise ValueError(f"{cls.__name__} is expected to be prefix:suffix, got {entity_string}")

    @classmethod
    def from_list(cls, entity_strings: list[str], base_prefix: str | None = None, **kwargs) -> list[Self]:
        return [
            cls.from_string(entity_string=entity_string, base_prefix=base_prefix, **kwargs)
            for entity_string in entity_strings
        ]


class ParentClass(Entity):
    type_: EntityTypes = EntityTypes.class_

    @property
    def view_id(self) -> ViewId:
        return ViewId(space=self.space, external_id=self.external_id, version=self.version)

    @classmethod
    def from_view_id(cls, view_id: ViewId) -> Self:
        return cls(prefix=view_id.space, suffix=view_id.external_id, version=view_id.version)


class ContainerEntity(Entity):
    type_: EntityTypes = EntityTypes.container


class ViewEntity(Entity):
    type_: EntityTypes = EntityTypes.view


class Triple(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    subject: str | URIRef | Entity
    predicate: str | URIRef | Entity
    object: str | URIRef | Literal | Entity | None = None
    optional: bool = Field(
        description="Indicates whether a triple is optional, used when building SPARQL query",
        default=False,
    )

    @classmethod
    def from_rdflib_triple(cls, triple: tuple[URIRef, URIRef, URIRef | Literal]) -> Self:
        return cls(subject=triple[0], predicate=triple[1], object=triple[2])



import itertools
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Set
from dataclasses import dataclass
from typing import Generic, TypeVar

import pandas as pd
from pydantic import BaseModel

from cognite.neat.rules.models._base import BaseRules
from cognite.neat.rules.models._rdfpath import RDFPath
from cognite.neat.rules.models.entities import (
    ClassEntity,
    Entity,
    ReferenceEntity,
)
from cognite.neat.rules.models.information import InformationProperty
from cognite.neat.utils.rdf_ import get_inheritance_path

T_Rules = TypeVar("T_Rules", bound=BaseRules)
T_Property = TypeVar("T_Property", bound=BaseModel)
T_Class = TypeVar("T_Class", bound=BaseModel)
T_ClassEntity = TypeVar("T_ClassEntity", bound=Entity)
T_PropertyEntity = TypeVar("T_PropertyEntity", bound=Entity | str)


@dataclass(frozen=True)
class Linkage(Generic[T_ClassEntity, T_PropertyEntity]):
    source_class: T_ClassEntity
    connecting_property: T_PropertyEntity
    target_class: T_ClassEntity
    max_occurrence: int | float | None


class LinkageSet(set, Generic[T_ClassEntity, T_PropertyEntity], Set[Linkage[T_ClassEntity, T_PropertyEntity]]):
    @property
    def source_class(self) -> set[T_ClassEntity]:
        return {link.source_class for link in self}

    @property
    def target_class(self) -> set[T_ClassEntity]:
        return {link.target_class for link in self}

    def get_target_classes_by_source(self) -> dict[T_ClassEntity, set[T_ClassEntity]]:
        target_classes_by_source: dict[T_ClassEntity, set[T_ClassEntity]] = defaultdict(set)
        for link in self:
            target_classes_by_source[link.source_class].add(link.target_class)
        return target_classes_by_source

    def to_pandas(self) -> pd.DataFrame:
        # Todo: Remove this method
        return pd.DataFrame(
            [
                {
                    "source_class": link.source_class,
                    "connecting_property": link.connecting_property,
                    "target_class": link.target_class,
                    "max_occurrence": link.max_occurrence,
                }
                for link in self
            ]
        )


class BaseAnalysis(ABC, Generic[T_Rules, T_Class, T_Property, T_ClassEntity, T_PropertyEntity]):
    def __init__(self, rules: T_Rules) -> None:
        self.rules = rules

    @abstractmethod
    def _get_classes(self) -> list[T_Class]:
        raise NotImplementedError

    @abstractmethod
    def _get_properties(self) -> list[T_Property]:
        raise NotImplementedError

    @abstractmethod
    def _get_reference(self, class_or_property: T_Class | T_Property) -> ReferenceEntity | None:
        raise NotImplementedError

    @abstractmethod
    def _get_cls_entity(self, class_: T_Class | T_Property) -> T_ClassEntity:
        raise NotImplementedError

    @abstractmethod
    def _get_prop_entity(self, property_: T_Property) -> T_PropertyEntity:
        raise NotImplementedError

    @abstractmethod
    def _get_cls_parents(self, class_: T_Class) -> list[T_ClassEntity] | None:
        raise NotImplementedError

    @abstractmethod
    def _get_reference_rules(self) -> T_Rules | None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _set_cls_entity(cls, property_: T_Property, class_: T_ClassEntity) -> None:
        raise NotImplementedError

    @abstractmethod
    def _get_object(self, property_: T_Property) -> T_ClassEntity | None:
        raise NotImplementedError

    @abstractmethod
    def _get_max_occurrence(self, property_: T_Property) -> int | float | None:
        raise NotImplementedError

    @property
    def directly_referred_classes(self) -> set[ClassEntity]:
        ref_rules = self._get_reference_rules()
        if ref_rules is None:
            return set()
        prefix = ref_rules.metadata.get_prefix()
        return {
            ref.as_class_entity()
            for class_ in self._get_classes()
            if isinstance((ref := self._get_reference(class_)), ReferenceEntity) and ref.prefix == prefix
        }

    @property
    def inherited_referred_classes(self) -> set[ClassEntity]:
        dir_referred_classes = self.directly_referred_classes
        inherited_referred_classes = []
        for class_ in dir_referred_classes:
            inherited_referred_classes.extend(self.class_inheritance_path(class_))
        return set(inherited_referred_classes)

    # Todo Lru cache this method.
    def class_parent_pairs(self) -> dict[T_ClassEntity, list[T_ClassEntity]]:
        """This only returns class - parent pairs only if parent is in the same data model"""
        class_subclass_pairs: dict[T_ClassEntity, list[T_ClassEntity]] = {}
        for cls_ in self._get_classes():
            entity = self._get_cls_entity(cls_)
            class_subclass_pairs[entity] = []
            for parent in self._get_cls_parents(cls_) or []:
                if parent.prefix == entity.prefix:
                    class_subclass_pairs[entity].append(parent)
                else:
                    warnings.warn(
                        f"Parent class {parent} of class {cls_} is not in the same namespace, skipping !",
                        stacklevel=2,
                    )

        return class_subclass_pairs

    def classes_with_properties(self, consider_inheritance: bool = False) -> dict[T_ClassEntity, list[T_Property]]:
        """Returns classes that have been defined in the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Dictionary of classes with a list of properties defined for them

        !!! note "consider_inheritance"
            If consider_inheritance is True, properties from parent classes will also be considered.
            This means if a class has a parent class, and the parent class has properties defined for it,
            while we do not have any properties defined for the child class, we will still consider the
            properties from the parent class. If consider_inheritance is False, we will only consider
            properties defined for the child class, thus if no properties are defined for the child class,
            it will not be included in the returned dictionary.
        """

        class_property_pairs: dict[T_ClassEntity, list[T_Property]] = defaultdict(list)

        for property_ in self._get_properties():
            class_property_pairs[self._get_cls_entity(property_)].append(property_)  # type: ignore

        if consider_inheritance:
            class_parent_pairs = self.class_parent_pairs()
            for class_ in class_parent_pairs:
                self._add_inherited_properties(class_, class_property_pairs, class_parent_pairs)

        return class_property_pairs

    def class_inheritance_path(self, class_: ClassEntity) -> list[ClassEntity]:
        class_parent_pairs = self.class_parent_pairs()
        return get_inheritance_path(class_, class_parent_pairs)

    @classmethod
    def _add_inherited_properties(
        cls,
        class_: T_ClassEntity,
        class_property_pairs: dict[T_ClassEntity, list[T_Property]],
        class_parent_pairs: dict[T_ClassEntity, list[T_ClassEntity]],
    ):
        inheritance_path = get_inheritance_path(class_, class_parent_pairs)
        for parent in inheritance_path:
            # ParentClassEntity -> ClassEntity to match the type of class_property_pairs
            if parent in class_property_pairs:
                for property_ in class_property_pairs[parent]:
                    property_ = property_.model_copy()

                    # This corresponds to importing properties from parent class
                    # making sure that the property is attached to desired child class
                    cls._set_cls_entity(property_, class_)

                    # need same if we have RDF path to make sure that the starting class is the
                    if class_ in class_property_pairs:
                        class_property_pairs[class_].append(property_)
                    else:
                        class_property_pairs[class_] = [property_]

    def class_property_pairs(
        self, only_rdfpath: bool = False, consider_inheritance: bool = False
    ) -> dict[T_ClassEntity, dict[T_PropertyEntity, T_Property]]:
        """Returns a dictionary of classes with a dictionary of properties associated with them.

        Args:
            only_rdfpath : To consider only properties which have rule `rdfpath` set. Defaults False
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Dictionary of classes with a dictionary of properties associated with them.

        !!! note "difference to get_classes_with_properties"
            This method returns a dictionary of classes with a dictionary of properties associated with them.
            While get_classes_with_properties returns a dictionary of classes with a list of
            properties defined for them,
            here we filter the properties based on the `only_rdfpath` parameter and only consider
            the first definition of a property if it is defined more than once.

        !!! note "only_rdfpath"
            If only_rdfpath is True, only properties with RuleType.rdfpath will be returned as
            a part of the dictionary of properties related to a class. Otherwise, all properties
            will be returned.

        !!! note "consider_inheritance"
            If consider_inheritance is True, properties from parent classes will also be considered.
            This means if a class has a parent class, and the parent class has properties defined for it,
            while we do not have any properties defined for the child class, we will still consider the
            properties from the parent class. If consider_inheritance is False, we will only consider
            properties defined for the child class, thus if no properties are defined for the child class,
            it will not be included in the returned dictionary.
        """
        # TODO: https://cognitedata.atlassian.net/jira/software/projects/NEAT/boards/893?selectedIssue=NEAT-78

        class_property_pairs: dict[T_ClassEntity, dict[T_PropertyEntity, T_Property]] = {}

        for class_, properties in self.classes_with_properties(consider_inheritance).items():
            processed_properties: dict[T_PropertyEntity, T_Property] = {}
            for property_ in properties:
                prop_entity = self._get_prop_entity(property_)
                if prop_entity in processed_properties:
                    # TODO: use appropriate Warning class from _exceptions.py
                    # if missing make one !
                    warnings.warn(
                        f"Property {processed_properties} for {class_} has been defined more than once!"
                        " Only the first definition will be considered, skipping the rest..",
                        stacklevel=2,
                    )
                    continue

                if (
                    only_rdfpath
                    and isinstance(property_, InformationProperty)
                    and isinstance(property_.transformation, RDFPath)
                ) or not only_rdfpath:
                    processed_properties[prop_entity] = property_
            class_property_pairs[class_] = processed_properties

        return class_property_pairs

    def class_linkage(self, consider_inheritance: bool = False) -> LinkageSet[T_ClassEntity, T_PropertyEntity]:
        """Returns a set of class linkages in the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:

        """
        class_linkage = LinkageSet[T_ClassEntity, T_PropertyEntity]()

        class_property_pairs = self.classes_with_properties(consider_inheritance)
        properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

        for property_ in properties:
            object_ = self._get_object(property_)
            if object_ is not None:
                class_linkage.add(
                    Linkage(
                        source_class=self._get_cls_entity(property_),
                        connecting_property=self._get_prop_entity(property_),
                        target_class=object_,
                        max_occurrence=self._get_max_occurrence(property_),
                    )
                )

        return class_linkage

    def connected_classes(self, consider_inheritance: bool = False) -> set[T_ClassEntity]:
        """Return a set of classes that are connected to other classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that are connected to other classes
        """
        class_linkage = self.class_linkage(consider_inheritance)
        return class_linkage.source_class.union(class_linkage.target_class)

    def defined_classes(self, consider_inheritance: bool = False) -> set[T_ClassEntity]:
        """Returns classes that have properties defined for them in the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that have been defined in the data model
        """
        class_property_pairs = self.classes_with_properties(consider_inheritance)
        properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

        return {self._get_cls_entity(property) for property in properties}

    def disconnected_classes(self, consider_inheritance: bool = False) -> set[T_ClassEntity]:
        """Return a set of classes that are disconnected (i.e. isolated) from other classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that are disconnected from other classes
        """
        return self.defined_classes(consider_inheritance) - self.connected_classes(consider_inheritance)

    def symmetrically_connected_classes(
        self, consider_inheritance: bool = False
    ) -> set[tuple[ClassEntity, ClassEntity]]:
        """Returns a set of pairs of symmetrically linked classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of pairs of symmetrically linked classes

        !!! note "Symmetrically Connected Classes"
            Symmetrically connected classes are classes that are connected to each other
            in both directions. For example, if class A is connected to class B, and class B
            is connected to class A, then classes A and B are symmetrically connected.
        """

        # TODO: Find better name for this method
        sym_pairs: set[tuple[ClassEntity, ClassEntity]] = set()

        class_linkage = self.class_linkage(consider_inheritance)
        if not class_linkage:
            return sym_pairs

        targets_by_source = class_linkage.get_target_classes_by_source()
        for link in class_linkage:
            source = link.source_class
            target = link.target_class

            if source in targets_by_source[source] and (source, target) not in sym_pairs:
                sym_pairs.add((source, target))
        return sym_pairs

    def as_property_dict(
        self,
    ) -> dict[T_PropertyEntity, list[T_Property]]:
        """This is used to capture all definitions of a property in the data model."""
        property_dict: dict[T_PropertyEntity, list[T_Property]] = defaultdict(list)
        for definition in self._get_properties():
            property_dict[self._get_prop_entity(definition)].append(definition)
        return property_dict

    def as_class_dict(self) -> dict[str, T_Class]:
        """This is to simplify access to classes through dict."""
        class_dict: dict[str, T_Class] = {}
        for definition in self._get_classes():
            entity = self._get_cls_entity(definition)
            if entity.suffix in class_dict:
                warnings.warn(
                    f"Class {entity} has been defined more than once! Only the first definition "
                    "will be considered, skipping the rest..",
                    stacklevel=2,
                )
                continue
            class_dict[entity.suffix] = definition
        return class_dict

    @abstractmethod
    def subset_rules(self, desired_classes: set[T_ClassEntity]) -> T_Rules:
        raise NotImplementedError



from abc import ABC
from dataclasses import dataclass
from typing import Any

from pydantic_core import ErrorDetails

from cognite.neat.issues import MultiValueError, NeatError, NeatIssue, NeatIssueList, NeatWarning

__all__ = [
    "ValidationIssue",
    "NeatValidationError",
    "DefaultPydanticError",
    "ValidationWarning",
    "IssueList",
    "MultiValueError",
]


@dataclass(frozen=True)
class ValidationIssue(NeatIssue, ABC): ...


@dataclass(frozen=True)
class NeatValidationError(NeatError, ValidationIssue, ABC):
    @classmethod
    def from_pydantic_errors(cls, errors: list[ErrorDetails], **kwargs) -> "list[NeatValidationError]":
        """Convert a list of pydantic errors to a list of Error instances.

        This is intended to be overridden in subclasses to handle specific error types.
        """
        all_errors: list[NeatValidationError] = []
        for error in errors:
            if isinstance(ctx := error.get("ctx"), dict) and isinstance(
                multi_error := ctx.get("error"), MultiValueError
            ):
                all_errors.extend(multi_error.errors)  # type: ignore[arg-type]
            else:
                all_errors.append(DefaultPydanticError.from_pydantic_error(error))
        return all_errors


@dataclass(frozen=True)
class DefaultPydanticError(NeatValidationError):
    type: str
    loc: tuple[int | str, ...]
    msg: str
    input: Any
    ctx: dict[str, Any] | None

    @classmethod
    def from_pydantic_error(cls, error: ErrorDetails) -> "NeatValidationError":
        return cls(
            type=error["type"],
            loc=error["loc"],
            msg=error["msg"],
            input=error.get("input"),
            ctx=error.get("ctx"),
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["type"] = self.type
        output["loc"] = self.loc
        output["msg"] = self.msg
        output["input"] = self.input
        output["ctx"] = self.ctx
        return output

    def message(self) -> str:
        if self.loc and len(self.loc) == 1:
            return f"{self.loc[0]} sheet: {self.msg}"
        elif self.loc and len(self.loc) == 2:
            return f"{self.loc[0]} sheet field/column <{self.loc[1]}>: {self.msg}"
        else:
            return self.msg


@dataclass(frozen=True)
class ValidationWarning(NeatWarning, ValidationIssue, ABC): ...


class IssueList(NeatIssueList[ValidationIssue]): ...



import logging
import warnings
from collections.abc import Hashable
from typing import Any, cast, no_type_check
from warnings import warn

import pandas as pd
from pydantic import field_validator
from pydantic_core import ErrorDetails, ValidationError
from rdflib import Namespace

from cognite.neat.constants import get_default_prefixes
from cognite.neat.exceptions import wrangle_warnings

# rules model and model components:
from cognite.neat.legacy.rules.models.rules import (
    Class,
    Metadata,
    Property,
    RuleModel,
    Rules,
)
from cognite.neat.legacy.rules.models.tables import Tables

# importers:
from cognite.neat.rules import exceptions
from cognite.neat.utils.auxiliary import generate_exception_report

__all__ = ["RawRules"]


class RawRules(RuleModel):
    """
    RawRules represent invalidated form of Rules, which is a core concept in `neat`.
    RawRules are used as staging area for Rules, and are often used when importing rules
    from sources other than Excel rules, e.g. from a json schema or owl ontology. Often
    these sources are not validated, and are missing information to be able to create
    a valid Rules object.

    Args:
        Metadata: Data model metadata
        classes: Classes defined in the data model
        properties: Class properties defined in the data model with accompanying transformation rules
                    to transform data from source to target representation
        prefixes: Prefixes used in the data model. Defaults to internal set of prefixes
        instances: Instances defined in the data model. Defaults to None
    """

    Metadata: pd.DataFrame
    Classes: pd.DataFrame
    Properties: pd.DataFrame
    Prefixes: pd.DataFrame = pd.DataFrame()
    Instances: pd.DataFrame = pd.DataFrame()
    importer_type: str = "RawTablesImporter"

    @field_validator("Metadata")
    def has_metadata_mandatory_rows(cls, v: pd.DataFrame):
        given_rows = set(v.iloc[:, 0].values)
        mandatory_rows = Metadata.mandatory_fields()
        mandatory_rows_alias = Metadata.mandatory_fields(use_alias=True)

        if not (mandatory_rows.issubset(given_rows) or mandatory_rows_alias.issubset(given_rows)):
            missing_rows = mandatory_rows_alias.difference(given_rows)
            raise exceptions.MetadataSheetMissingMandatoryFields(missing_rows).to_pydantic_custom_error()
        return v

    @field_validator("Classes")
    def has_classes_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = Class.mandatory_fields()
        mandatory_columns_alias = Class.mandatory_fields(use_alias=True)

        if not (mandatory_columns.issubset(given_columns) or mandatory_columns_alias.issubset(given_columns)):
            missing_columns = mandatory_columns_alias.difference(given_columns)
            raise exceptions.ClassesSheetMissingMandatoryColumns(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Properties")
    def has_properties_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = Property.mandatory_fields()
        mandatory_columns_alias = Property.mandatory_fields(use_alias=True)

        if not (mandatory_columns.issubset(given_columns) or mandatory_columns_alias.issubset(given_columns)):
            missing_columns = mandatory_columns_alias.difference(given_columns)
            raise exceptions.PropertiesSheetMissingMandatoryColumns(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Prefixes")
    def has_prefixes_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = {"Prefix", "URI"}

        if not mandatory_columns.issubset(given_columns):
            missing_columns = mandatory_columns.difference(given_columns)
            raise exceptions.PrefixesSheetMissingMandatoryColumns(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Instances")
    def has_instances_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = {"Instance", "Property", "Value"}

        if not mandatory_columns.issubset(given_columns):
            missing_columns = mandatory_columns.difference(given_columns)
            raise exceptions.InstancesSheetMissingMandatoryColumns(missing_columns).to_pydantic_custom_error()
        return v

    @staticmethod
    def _drop_non_string_columns(df: pd.DataFrame) -> pd.DataFrame:
        """Drop non-string columns as this can cause issue when loading rules

        Args:
            df: data frame

        Returns:
            dataframe with removed non string columns
        """
        columns = [column for column in df.columns[df.columns.notna()] if isinstance(column, str)]

        return df[columns]

    # mypy complains "RawRules" has incompatible type "**dict[str, DataFrame]"; expected "set[str]" , which is wrong!
    @no_type_check
    @classmethod
    def from_tables(cls, tables: dict[str, pd.DataFrame], importer_type: str = "RawTablesImporter") -> "RawRules":
        """Create RawRules from raw tables.

        Args:
            tables: Tables to be converted to RawRules

        Returns:
            Instance of RawRules
        """

        expected_tables = cls.mandatory_fields()

        # Validate raw tables
        if missing_tables := (expected_tables - set(tables)):
            raise exceptions.SourceObjectDoesNotProduceMandatorySheets(missing_tables)

        tables_dict: dict[str, pd.DataFrame] = {
            Tables.metadata: tables[Tables.metadata],
            Tables.classes: cls._drop_non_string_columns(tables[Tables.classes]),
            Tables.properties: cls._drop_non_string_columns(tables[Tables.properties]),
        }

        if Tables.prefixes in tables:
            tables_dict[Tables.prefixes] = cls._drop_non_string_columns(tables[Tables.prefixes])
        if Tables.instances in tables:
            tables_dict[Tables.instances] = cls._drop_non_string_columns(tables[Tables.instances])

        return cls(
            **tables_dict,
            importer_type=importer_type,
        )

    # mypy unsatisfied with overload , tried all combination and gave up
    @no_type_check
    def to_rules(
        self,
        return_report: bool = False,
        skip_validation: bool = False,
        validators_to_skip: set[str] | None = None,
    ) -> tuple[Rules | None, list[ErrorDetails] | None, list | None] | Rules:
        """Validates RawRules instances and returns Rules instance.

        Args:
            return_report: To return validation report. Defaults to False.
            skip_validation: Bypasses Rules validation. Defaults to False.
            validators_to_skip: List of validators to skip. Defaults to None.

        Returns:
            Instance of `Rules`, which can be validated or not validated based on
            `skip_validation` flag, and optional list of errors and warnings if
            `return_report` is set to True.

        !!! Note
            `skip_validation` flag should be only used for purpose when `Rules` object
            is exported to an Excel file. Do not use this flag for any other purpose!
        """

        rules_dict = _raw_tables_to_rules_dict(self, validators_to_skip)
        if skip_validation:
            return _to_invalidated_rules(rules_dict)
        else:
            return _to_validated_rules(rules_dict, return_report)

    def validate_rules(self) -> str | None:
        _, errors, warnings_ = self.to_rules(return_report=True, skip_validation=False)

        report = ""
        if errors:
            warnings.warn(
                exceptions.RulesHasErrors(importer_type=self.importer_type).message,
                category=exceptions.RulesHasErrors,
                stacklevel=2,
            )
            report = generate_exception_report(cast(list[ErrorDetails], errors), "Errors")

        if warnings_:
            warnings.warn(
                exceptions.RulesHasWarnings(importer_type=self.importer_type).message,
                category=exceptions.RulesHasWarnings,
                stacklevel=2,
            )
            report += generate_exception_report(cast(list[dict], warnings_), "Warnings")

        return report if report else None


def _to_validated_rules(
    rules_dict: dict, return_report: bool = False
) -> tuple[Rules | None, list[ErrorDetails] | None, list[dict] | None] | Rules:
    validation_warnings = []
    try:
        with warnings.catch_warnings(record=True) as validation_warnings:
            rules = Rules(**rules_dict)
        return (rules, None, wrangle_warnings(validation_warnings)) if return_report else rules

    except ValidationError as e:
        validation_errors = e.errors()
        if return_report:
            return None, validation_errors, wrangle_warnings(validation_warnings)
        else:
            raise e


def _to_invalidated_rules(rules_dict: dict) -> Rules:
    args = {
        "metadata": Metadata.model_construct(**rules_dict["metadata"]),
        "classes": {
            class_: Class.model_construct(**definition) for class_, definition in rules_dict["classes"].items()
        },
        "properties": {
            property_: Property.model_construct(**definition)
            for property_, definition in rules_dict["properties"].items()
        },
        "prefixes": rules_dict["prefixes"],
    }

    return cast(Rules, Rules.model_construct(**args))


def _raw_tables_to_rules_dict(raw_tables: RawRules, validators_to_skip: set | None = None) -> dict[str, Any]:
    """Converts raw tables to a dictionary of rules."""
    rules_dict: dict[str, Any] = {
        "metadata": _metadata_table2dict(raw_tables.Metadata),
        "classes": _classes_table2dict(raw_tables.Classes),
        "properties": _properties_table2dict(raw_tables.Properties),
        "prefixes": (
            get_default_prefixes() if raw_tables.Prefixes.empty else _prefixes_table2dict(raw_tables.Prefixes)
        ),
    }

    rules_dict["instances"] = (
        None
        if raw_tables.Instances.empty
        else _instances_table2dict(raw_tables.Instances, rules_dict["metadata"], rules_dict["prefixes"])
    )

    if validators_to_skip:
        rules_dict["validators_to_skip"] = validators_to_skip
        rules_dict["metadata"]["validators_to_skip"] = validators_to_skip
        for class_ in rules_dict["classes"].keys():
            rules_dict["classes"][class_]["validators_to_skip"] = validators_to_skip
        for property_ in rules_dict["properties"].keys():
            rules_dict["properties"][property_]["validators_to_skip"] = validators_to_skip

    return rules_dict


def _metadata_table2dict(meta_df: pd.DataFrame) -> dict[str, Any]:
    assert len(meta_df.columns) == 2
    col1, col2 = meta_df.columns
    metadata_dict = dict(zip(meta_df[col1], meta_df[col2], strict=True))
    metadata_dict["source"] = meta_df.source if "source" in dir(meta_df) else None
    if "namespace" in metadata_dict:
        metadata_dict["namespace"] = Namespace(metadata_dict["namespace"])
    return metadata_dict


def _classes_table2dict(
    classes_df: pd.DataFrame,
) -> dict[Any | None, dict[Hashable, Any]]:
    return {class_.get("Class"): class_ for class_ in classes_df.to_dict(orient="records")}


def _properties_table2dict(
    properties_df: pd.DataFrame,
) -> dict[str, dict[Hashable, Any]]:
    return {f"row {i+3}": property_ for i, property_ in enumerate(properties_df.to_dict(orient="records"))}


def _prefixes_table2dict(prefix_df: pd.DataFrame) -> dict[str, Namespace]:
    return {row["Prefix"]: Namespace(row["URI"]) for i, row in prefix_df.iterrows()}


def _instances_table2dict(
    instances_df: pd.DataFrame, metadata: dict[str, Any], prefixes: dict[str, Namespace]
) -> list[dict] | None:
    if ("prefix" not in metadata and "namespace" not in metadata) or "namespace" not in metadata:
        logging.warning(exceptions.MissingDataModelPrefixOrNamespace().message)
        warn(exceptions.MissingDataModelPrefixOrNamespace().message, stacklevel=2)
        return None

    prefix = metadata["prefix"] if "prefix" in metadata else metadata["space"]
    prefixes[prefix] = metadata["namespace"]

    instances = []
    for _, row in instances_df.iterrows():
        row_as_dict = row.to_dict()
        row_as_dict["namespace"] = metadata["namespace"]
        row_as_dict["prefixes"] = prefixes
        instances.append(row_as_dict)
    return instances



import logging
import os
from pathlib import Path

import yaml


def migrate_wf_manifest(wf_store_path: Path):
    """Migrate workflow manifests . Changed name of element from groups -> system_components"""
    migrated_files = []
    wf_store_path = wf_store_path / "workflows"
    for wf_module_name in os.listdir(wf_store_path):
        wf_module_full_path = wf_store_path / wf_module_name

        if wf_module_full_path.is_dir():
            metadata_file = wf_store_path / wf_module_name / "workflow.yaml"
            metadata_file_migrated = wf_store_path / wf_module_name / "workflow.yaml"
            logging.info(f"Loading workflow {wf_module_name} metadata from {metadata_file}")
            if metadata_file.exists():
                with metadata_file.open() as f:
                    manifest_yaml = yaml.safe_load(f)
                    logging.info(f"Loaded workflow {wf_module_name} metadata from {metadata_file}")
                if "groups" in manifest_yaml:
                    logging.info(f"Found groups in {metadata_file}, migrating to system_components")
                    manifest_yaml["system_components"] = manifest_yaml["groups"]
                    del manifest_yaml["groups"]
                    with metadata_file_migrated.open("w") as f:
                        yaml.dump(manifest_yaml, f, indent=4)
                        migrated_files.append(metadata_file_migrated)
            else:
                logging.info(f"Metadata file {metadata_file} not found, skipping")
                continue
    return migrated_files



from abc import abstractmethod
from collections.abc import Iterable

from cognite.neat.graph.models import Triple
from cognite.neat.utils.auxiliary import class_html_doc


class BaseExtractor:
    """This is the base class for all extractors. It defines the interface that
    extractors must implement.
    """

    @abstractmethod
    def extract(self) -> Iterable[Triple]:
        raise NotImplementedError()

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)



from pathlib import Path

from rdflib import Graph, Namespace

from cognite.neat.constants import get_default_prefixes


def rdf_file_to_graph(
    graph: Graph,
    filepath: Path,
    base_prefix: str | None = None,
    base_namespace: Namespace | None = None,
    prefixes: dict[str, Namespace] | None = None,
) -> Graph:
    """Created rdflib Graph instance loaded with RDF triples from file

    Args:
        filepath: Path to the RDF file
        base_prefix: base prefix for URIs. Defaults to None.
        base_namespace: base namespace for URIs . Defaults to None.
        prefixes: Dictionary of prefixes to bind to graph. Defaults to internal set of prefixes.
        graph: Graph instance to load RDF triples into. Defaults to None.

    Returns:
        Graph instance loaded with RDF triples from file
    """

    prefixes = prefixes if prefixes else get_default_prefixes()

    if filepath.is_file():
        graph.parse(filepath, publicID=base_namespace)
    else:
        for filename in filepath.iterdir():
            if filename.is_file():
                graph.parse(filename, publicID=base_namespace)
    if base_prefix and base_namespace:
        graph.bind(base_prefix, base_namespace)
    if prefixes:
        for prefix, namespace in prefixes.items():
            graph.bind(prefix, namespace)

    return graph



import webbrowser
from pathlib import Path

import uvicorn
from fastapi import FastAPI

app = FastAPI()


def run():
    (Path.cwd() / "data").mkdir(parents=True, exist_ok=True)
    webbrowser.open("http://localhost:8000/")
    uvicorn.run("cognite.neat.app.api.explorer:app", host="0.0.0.0", port=8000, reload=False)


if __name__ == "__main__":
    run()



import math
import re
import sys
import warnings
from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from cognite.client import data_modeling as dm
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic.main import IncEx
from pydantic_core.core_schema import ValidationInfo

from cognite.neat.rules import issues
from cognite.neat.rules.issues import MultiValueError
from cognite.neat.rules.models._base import (
    BaseMetadata,
    BaseRules,
    DataModelType,
    ExtensionCategory,
    RoleTypes,
    SchemaCompleteness,
    SheetEntity,
    SheetList,
)
from cognite.neat.rules.models._types import (
    ExternalIdType,
    PropertyType,
    StrListType,
    VersionType,
)
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.domain import DomainRules
from cognite.neat.rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    ContainerEntityList,
    DMSUnknownEntity,
    ReferenceEntity,
    URLEntity,
    ViewEntity,
    ViewEntityList,
    ViewPropertyEntity,
)
from cognite.neat.rules.models.wrapped_entities import HasDataFilter, NodeTypeFilter, RawFilter

from ._schema import DMSSchema

if TYPE_CHECKING:
    from cognite.neat.rules.models.information._rules import InformationRules

if sys.version_info >= (3, 11):
    pass
else:
    pass

_DEFAULT_VERSION = "1"


class DMSMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.dms
    data_model_type: DataModelType = Field(DataModelType.enterprise, alias="dataModelType")
    schema_: SchemaCompleteness = Field(alias="schema")
    extension: ExtensionCategory = ExtensionCategory.addition
    space: ExternalIdType
    name: str | None = Field(
        None,
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, min_length=1, max_length=1024)
    external_id: ExternalIdType = Field(alias="externalId")
    version: VersionType
    creator: StrListType
    created: datetime = Field(
        description=("Date of the data model creation"),
    )
    updated: datetime = Field(
        description=("Date of the data model update"),
    )

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value

    @field_serializer("schema_", "extension", "data_model_type", when_used="always")
    @staticmethod
    def as_string(value: SchemaCompleteness | ExtensionCategory | DataModelType) -> str:
        return str(value)

    @field_validator("schema_", mode="plain")
    def as_enum_schema(cls, value: str) -> SchemaCompleteness:
        return SchemaCompleteness(value)

    @field_validator("extension", mode="plain")
    def as_enum_extension(cls, value: str) -> ExtensionCategory:
        return ExtensionCategory(value)

    @field_validator("data_model_type", mode="plain")
    def as_enum_model_type(cls, value: str) -> DataModelType:
        return DataModelType(value)

    @field_validator("description", mode="before")
    def nan_as_none(cls, value):
        if isinstance(value, float) and math.isnan(value):
            return None
        return value

    def as_space(self) -> dm.SpaceApply:
        return dm.SpaceApply(
            space=self.space,
        )

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    def as_data_model(self) -> dm.DataModelApply:
        suffix = f"Creator: {', '.join(self.creator)}"
        if self.description:
            description = f"{self.description} Creator: {', '.join(self.creator)}"
        else:
            description = suffix

        return dm.DataModelApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name or None,
            version=self.version or "missing",
            description=description,
            views=[],
        )

    def as_identifier(self) -> str:
        return repr(self.as_data_model_id())

    @classmethod
    def _get_description_and_creator(cls, description_raw: str | None) -> tuple[str | None, list[str]]:
        if description_raw and (description_match := re.search(r"Creator: (.+)", description_raw)):
            creator = description_match.group(1).split(", ")
            description = description_raw.replace(description_match.string, "").strip() or None
        elif description_raw:
            creator = ["MISSING"]
            description = description_raw
        else:
            creator = ["MISSING"]
            description = None
        return description, creator

    @classmethod
    def from_data_model(cls, data_model: dm.DataModelApply, has_reference: bool) -> "DMSMetadata":
        description, creator = cls._get_description_and_creator(data_model.description)
        return cls(
            schema_=SchemaCompleteness.complete,
            data_model_type=DataModelType.solution if has_reference else DataModelType.enterprise,
            space=data_model.space,
            name=data_model.name or None,
            description=description,
            external_id=data_model.external_id,
            version=data_model.version,
            creator=creator,
            created=datetime.now(),
            updated=datetime.now(),
        )

    def get_prefix(self) -> str:
        return self.space


class DMSProperty(SheetEntity):
    view: ViewEntity = Field(alias="View")
    view_property: str = Field(alias="View Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    connection: Literal["direct", "edge", "reverse"] | None = Field(None, alias="Connection")
    value_type: DataType | ViewPropertyEntity | ViewEntity | DMSUnknownEntity = Field(alias="Value Type")
    nullable: bool | None = Field(default=None, alias="Nullable")
    is_list: bool | None = Field(default=None, alias="Is List")
    default: str | int | dict | None = Field(None, alias="Default")
    reference: URLEntity | ReferenceEntity | None = Field(default=None, alias="Reference", union_mode="left_to_right")
    container: ContainerEntity | None = Field(None, alias="Container")
    container_property: str | None = Field(None, alias="Container Property")
    index: StrListType | None = Field(None, alias="Index")
    constraint: StrListType | None = Field(None, alias="Constraint")
    class_: ClassEntity = Field(alias="Class (linage)")
    property_: PropertyType = Field(alias="Property (linage)")

    @field_validator("nullable")
    def direct_relation_must_be_nullable(cls, value: Any, info: ValidationInfo) -> None:
        if info.data.get("connection") == "direct" and value is False:
            raise ValueError("Direct relation must be nullable")
        return value

    @field_validator("value_type", mode="after")
    def connections_value_type(
        cls, value: ViewPropertyEntity | ViewEntity | DMSUnknownEntity, info: ValidationInfo
    ) -> DataType | ViewPropertyEntity | ViewEntity | DMSUnknownEntity:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and not isinstance(value, ViewEntity | DMSUnknownEntity):
            raise ValueError(f"Direct relation must have a value type that points to a view, got {value}")
        elif connection == "edge" and not isinstance(value, ViewEntity):
            raise ValueError(f"Edge connection must have a value type that points to a view, got {value}")
        elif connection == "reverse" and not isinstance(value, ViewPropertyEntity | ViewEntity):
            raise ValueError(
                f"Reverse connection must have a value type that points to a view or view property, got {value}"
            )
        return value

    @field_serializer("value_type", when_used="always")
    @staticmethod
    def as_dms_type(value_type: DataType | ViewPropertyEntity | ViewEntity) -> str:
        if isinstance(value_type, DataType):
            return value_type.dms._type
        else:
            return str(value_type)


class DMSContainer(SheetEntity):
    container: ContainerEntity = Field(alias="Container")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    constraint: ContainerEntityList | None = Field(None, alias="Constraint")
    class_: ClassEntity = Field(alias="Class (linage)")

    def as_container(self) -> dm.ContainerApply:
        container_id = self.container.as_id()
        constraints: dict[str, dm.Constraint] = {}
        for constraint in self.constraint or []:
            requires = dm.RequiresConstraint(constraint.as_id())
            constraints[f"{constraint.space}_{constraint.external_id}"] = requires

        return dm.ContainerApply(
            space=container_id.space,
            external_id=container_id.external_id,
            name=self.name or None,
            description=self.description,
            constraints=constraints or None,
            properties={},
        )

    @classmethod
    def from_container(cls, container: dm.ContainerApply) -> "DMSContainer":
        constraints: list[ContainerEntity] = []
        for _, constraint_obj in (container.constraints or {}).items():
            if isinstance(constraint_obj, dm.RequiresConstraint):
                constraints.append(ContainerEntity.from_id(constraint_obj.require))
            # UniquenessConstraint it handled in the properties
        container_entity = ContainerEntity.from_id(container.as_id())
        return cls(
            class_=container_entity.as_class(),
            container=container_entity,
            name=container.name or None,
            description=container.description,
            constraint=constraints or None,
        )


class DMSView(SheetEntity):
    view: ViewEntity = Field(alias="View")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    implements: ViewEntityList | None = Field(None, alias="Implements")
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    filter_: HasDataFilter | NodeTypeFilter | RawFilter | None = Field(None, alias="Filter")
    in_model: bool = Field(True, alias="In Model")
    class_: ClassEntity = Field(alias="Class (linage)")

    def as_view(self) -> dm.ViewApply:
        view_id = self.view.as_id()
        implements = [parent.as_id() for parent in self.implements or []] or None
        if implements is None and isinstance(self.reference, ReferenceEntity):
            # Fallback to the reference if no implements are provided
            parent = self.reference.as_view_id()
            if (parent.space, parent.external_id) != (view_id.space, view_id.external_id):
                implements = [parent]

        return dm.ViewApply(
            space=view_id.space,
            external_id=view_id.external_id,
            version=view_id.version or _DEFAULT_VERSION,
            name=self.name or None,
            description=self.description,
            implements=implements,
            properties={},
        )

    @classmethod
    def from_view(cls, view: dm.ViewApply, in_model: bool) -> "DMSView":
        view_entity = ViewEntity.from_id(view.as_id())
        class_entity = view_entity.as_class(skip_version=True)

        return cls(
            class_=class_entity,
            view=view_entity,
            description=view.description,
            name=view.name,
            implements=[ViewEntity.from_id(parent, _DEFAULT_VERSION) for parent in view.implements] or None,
            in_model=in_model,
        )


class DMSRules(BaseRules):
    metadata: DMSMetadata = Field(alias="Metadata")
    properties: SheetList[DMSProperty] = Field(alias="Properties")
    views: SheetList[DMSView] = Field(alias="Views")
    containers: SheetList[DMSContainer] | None = Field(None, alias="Containers")
    last: "DMSRules | None" = Field(None, alias="Last", description="The previous version of the data model")
    reference: "DMSRules | None" = Field(None, alias="Reference")

    @field_validator("reference")
    def check_reference_of_reference(cls, value: "DMSRules | None", info: ValidationInfo) -> "DMSRules | None":
        if value is None:
            return None
        if value.reference is not None:
            raise ValueError("Reference rules cannot have a reference")
        if value.metadata.data_model_type == DataModelType.solution and (metadata := info.data.get("metadata")):
            warnings.warn(
                issues.dms.SolutionOnTopOfSolutionModelWarning(
                    metadata.as_data_model_id(), value.metadata.as_data_model_id()
                ),
                stacklevel=2,
            )
        return value

    @field_validator("views")
    def matching_version_and_space(cls, value: SheetList[DMSView], info: ValidationInfo) -> SheetList[DMSView]:
        if not (metadata := info.data.get("metadata")):
            return value
        model_version = metadata.version
        if different_version := [view.view.as_id() for view in value if view.view.version != model_version]:
            warnings.warn(issues.dms.ViewModelVersionNotMatchingWarning(different_version, model_version), stacklevel=2)
        if different_space := [view.view.as_id() for view in value if view.view.space != metadata.space]:
            warnings.warn(issues.dms.ViewModelSpaceNotMatchingWarning(different_space, metadata.space), stacklevel=2)
        return value

    @model_validator(mode="after")
    def post_validation(self) -> "DMSRules":
        from ._validation import DMSPostValidation

        issue_list = DMSPostValidation(self).validate()
        if issue_list.warnings:
            issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError([error for error in issue_list if isinstance(error, issues.NeatValidationError)])
        return self

    def dump(
        self,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        as_reference: bool = False,
    ) -> dict[str, Any]:
        from ._serializer import _DMSRulesSerializer

        dumped = self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        space, version = self.metadata.space, self.metadata.version
        serializer = _DMSRulesSerializer(by_alias, space, version)
        clean = serializer.clean(dumped, as_reference)
        last = "Last" if by_alias else "last"
        if last_dump := clean.get(last):
            clean[last] = serializer.clean(last_dump, False)
        reference = "Reference" if by_alias else "reference"
        if self.reference and (ref_dump := clean.get(reference)):
            space, version = self.reference.metadata.space, self.reference.metadata.version
            clean[reference] = _DMSRulesSerializer(by_alias, space, version).clean(ref_dump, True)
        return clean

    def as_schema(self, include_pipeline: bool = False, instance_space: str | None = None) -> DMSSchema:
        from ._exporter import _DMSExporter

        return _DMSExporter(self, include_pipeline, instance_space).to_schema()

    def as_information_rules(self) -> "InformationRules":
        from ._converter import _DMSRulesConverter

        return _DMSRulesConverter(self).as_information_rules()

    def as_domain_expert_rules(self) -> DomainRules:
        from ._converter import _DMSRulesConverter

        return _DMSRulesConverter(self).as_domain_rules()

    def create_reference(
        self, reference: "DMSRules", view_extension_mapping: dict[str, str], default_extension: str | None = None
    ) -> None:
        """Takes this data models and makes it into an extension of the reference data model.

        The argument view_extension_mapping is a dictionary where the keys are views of this data model,
        and each value is the view of the reference data model that the view should extend. For example:

        ```python
        view_extension_mapping = {"Pump": "Asset"}
        ```

        This would make the view "Pump" in this data model extend the view "Asset" in the reference data model.
        Note that all the keys in the dictionary must be external ids of views in this data model,
        and all the values must be external ids of views in the reference data model.

        Args:
            reference: The reference data model
            view_extension_mapping: A dictionary mapping views in this data model to views in the reference data model
            default_extension: The default view in the reference data model that views in this
                data model should extend if no mapping is provided.

        """
        if self.reference is not None:
            raise ValueError("Reference already exists")
        self.reference = reference
        view_by_external_id = {view.view.external_id: view for view in self.views}
        ref_view_by_external_id = {view.view.external_id: view for view in reference.views}

        if invalid_views := set(view_extension_mapping.keys()) - set(view_by_external_id.keys()):
            raise ValueError(f"Views are not in this dat model {invalid_views}")
        if invalid_views := set(view_extension_mapping.values()) - set(ref_view_by_external_id.keys()):
            raise ValueError(f"Views are not in the reference data model {invalid_views}")
        if default_extension and default_extension not in ref_view_by_external_id:
            raise ValueError(f"Default extension view not in the reference data model {default_extension}")

        properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in self.properties:
            properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        ref_properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in reference.properties:
            ref_properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        for view_external_id, view in view_by_external_id.items():
            if view_external_id in view_extension_mapping:
                ref_external_id = view_extension_mapping[view_external_id]
            elif default_extension:
                ref_external_id = default_extension
            else:
                continue

            ref_view = ref_view_by_external_id[ref_external_id]
            shared_properties = set(properties_by_view_external_id[view_external_id].keys()) & set(
                ref_properties_by_view_external_id[ref_external_id].keys()
            )
            if shared_properties:
                if view.implements is None:
                    view.implements = [ref_view.view]
                elif isinstance(view.implements, list) and ref_view.view not in view.implements:
                    view.implements.append(ref_view.view)
            for prop_name in shared_properties:
                prop = properties_by_view_external_id[view_external_id][prop_name]
                ref_prop = ref_properties_by_view_external_id[ref_external_id][prop_name]
                if ref_prop.container and ref_prop.container_property:
                    prop.container = ref_prop.container
                    prop.container_property = ref_prop.container_property
                prop.reference = ReferenceEntity.from_entity(ref_prop.view, ref_prop.view_property)



import json
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import TimeSeries, TimeSeriesList
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.auxiliary import string_to_ideal_type


class TimeSeriesExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusions TimeSeries into Neat.

    Args:
        timeseries (Iterable[TimeSeries]): An iterable of timeseries.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
    """

    def __init__(
        self,
        timeseries: Iterable[TimeSeries],
        namespace: Namespace | None = None,
        unpack_metadata: bool = True,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.timeseries = timeseries
        self.unpack_metadata = unpack_metadata

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        unpack_metadata: bool = True,
    ):
        return cls(
            cast(
                Iterable[TimeSeries],
                client.time_series(data_set_external_ids=data_set_external_id),
            ),
            namespace,
            unpack_metadata,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        namespace: Namespace | None = None,
        unpack_metadata: bool = True,
    ):
        return cls(TimeSeriesList.load(Path(file_path).read_text()), namespace, unpack_metadata)

    def extract(self) -> Iterable[Triple]:
        """Extract timeseries as triples."""
        for timeseries in self.timeseries:
            yield from self._timeseries2triples(timeseries)

    def _timeseries2triples(self, timeseries: TimeSeries) -> list[Triple]:
        id_ = self.namespace[f"TimeSeries_{timeseries.id}"]

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, self.namespace.TimeSeries)]

        # Create attributes

        if timeseries.external_id:
            triples.append((id_, self.namespace.external_id, Literal(timeseries.external_id)))

        if timeseries.name:
            triples.append((id_, self.namespace.name, Literal(timeseries.name)))

        if timeseries.is_string:
            triples.append((id_, self.namespace.is_string, Literal(timeseries.is_string)))

        if timeseries.metadata:
            if self.unpack_metadata:
                for key, value in timeseries.metadata.items():
                    if value:
                        type_aware_value = string_to_ideal_type(value)
                        try:
                            triples.append((id_, self.namespace[key], URIRef(str(AnyHttpUrl(type_aware_value)))))  # type: ignore
                        except ValidationError:
                            triples.append((id_, self.namespace[key], Literal(type_aware_value)))
            else:
                triples.append(
                    (
                        id_,
                        self.namespace.metadata,
                        Literal(json.dumps(timeseries.metadata)),
                    )
                )

        if timeseries.unit:
            triples.append((id_, self.namespace.unit, Literal(timeseries.unit)))

        if self.namespace.is_step:
            triples.append((id_, self.namespace.is_step, Literal(timeseries.is_step)))

        if timeseries.description:
            triples.append((id_, self.namespace.description, Literal(timeseries.description)))

        if timeseries.security_categories:
            for category in timeseries.security_categories:
                triples.append((id_, self.namespace.security_categories, Literal(category)))

        if timeseries.created_time:
            triples.append(
                (
                    id_,
                    self.namespace.created_time,
                    Literal(datetime.fromtimestamp(timeseries.created_time / 1000, timezone.utc)),
                )
            )

        if timeseries.last_updated_time:
            triples.append(
                (
                    id_,
                    self.namespace.last_updated_time,
                    Literal(datetime.fromtimestamp(timeseries.last_updated_time / 1000, timezone.utc)),
                )
            )

        if timeseries.legacy_name:
            triples.append((id_, self.namespace.legacy_name, Literal(timeseries.legacy_name)))

        # Create connections
        if timeseries.unit_external_id:
            # try to create connection to QUDT unit catalog
            try:
                triples.append(
                    (
                        id_,
                        self.namespace.unit_external_id,
                        URIRef(str(AnyHttpUrl(timeseries.unit_external_id))),
                    )
                )
            except ValidationError:
                triples.append(
                    (
                        id_,
                        self.namespace.unit_external_id,
                        Literal(timeseries.unit_external_id),
                    )
                )

        if timeseries.data_set_id:
            triples.append(
                (
                    id_,
                    self.namespace.dataset,
                    self.namespace[f"Dataset_{timeseries.data_set_id}"],
                )
            )

        if timeseries.asset_id:
            triples.append(
                (
                    id_,
                    self.namespace.asset,
                    self.namespace[f"Asset_{timeseries.asset_id}"],
                )
            )

        return triples



import logging
import warnings
from typing import Any, cast

from pydantic import ValidationError
from rdflib import URIRef

from cognite.neat.rules.models import SchemaCompleteness
from cognite.neat.rules.models._rdfpath import (
    Hop,
    RDFPath,
    SelfReferenceProperty,
    SingleProperty,
)
from cognite.neat.rules.models.entities import ClassEntity, ReferenceEntity
from cognite.neat.rules.models.information import (
    InformationClass,
    InformationProperty,
    InformationRules,
)
from cognite.neat.utils.rdf_ import get_inheritance_path

from ._base import BaseAnalysis


class InformationAnalysis(BaseAnalysis[InformationRules, InformationClass, InformationProperty, ClassEntity, str]):
    """Assumes analysis over only the complete schema"""

    def _get_object(self, property_: InformationProperty) -> ClassEntity | None:
        return property_.value_type if isinstance(property_.value_type, ClassEntity) else None

    def _get_max_occurrence(self, property_: InformationProperty) -> int | float | None:
        return property_.max_count

    def _get_reference(self, class_or_property: InformationClass | InformationProperty) -> ReferenceEntity | None:
        return class_or_property.reference if isinstance(class_or_property.reference, ReferenceEntity) else None

    def _get_cls_entity(self, class_: InformationClass | InformationProperty) -> ClassEntity:
        return class_.class_

    @classmethod
    def _set_cls_entity(cls, property_: InformationProperty, class_: ClassEntity) -> None:
        property_.class_ = class_

    def _get_prop_entity(self, property_: InformationProperty) -> str:
        return property_.property_

    def _get_cls_parents(self, class_: InformationClass) -> list[ClassEntity] | None:
        return list(class_.parent or []) or None

    def _get_reference_rules(self) -> InformationRules | None:
        return self.rules.reference

    def _get_properties(self) -> list[InformationProperty]:
        return list(self.rules.properties)

    def _get_classes(self) -> list[InformationClass]:
        return list(self.rules.classes)

    def has_hop_transformations(self):
        return any(
            prop_.transformation and isinstance(prop_.transformation.traversal, Hop) for prop_ in self.rules.properties
        )

    def has_self_reference_property_transformations(self):
        return any(
            prop_.transformation and isinstance(prop_.transformation.traversal, SelfReferenceProperty)
            for prop_ in self.rules.properties
        )

    def all_reference_transformations(self):
        return [
            prop_
            for prop_ in self.rules.properties
            if prop_.transformation and isinstance(prop_.transformation.traversal, SelfReferenceProperty)
        ]

    def define_property_renaming_config(self, class_: ClassEntity) -> dict[str | URIRef, str]:
        property_renaming_configuration: dict[str | URIRef, str] = {}

        if definitions := self.class_property_pairs(only_rdfpath=True, consider_inheritance=True).get(class_, None):
            for property_id, definition in definitions.items():
                transformation = cast(RDFPath, definition.transformation)

                # use case we have a single property rdf path, and defined prefix
                # in either metadata or prefixes of rules
                if isinstance(
                    transformation.traversal,
                    SingleProperty,
                ) and (
                    transformation.traversal.property.prefix in self.rules.prefixes
                    or transformation.traversal.property.prefix == self.rules.metadata.prefix
                ):
                    namespace = (
                        self.rules.metadata.namespace
                        if transformation.traversal.property.prefix == self.rules.metadata.prefix
                        else self.rules.prefixes[transformation.traversal.property.prefix]
                    )

                    property_renaming_configuration[namespace[transformation.traversal.property.suffix]] = property_id

                # otherwise we default to the property id
                else:
                    property_renaming_configuration[property_id] = property_id

        return property_renaming_configuration

    def subset_rules(self, desired_classes: set[ClassEntity]) -> InformationRules:
        """
        Subset rules to only include desired classes and their properties.

        Args:
            desired_classes: Desired classes to include in the reduced data model

        Returns:
            Instance of InformationRules

        !!! note "Inheritance"
            If desired classes contain a class that is a subclass of another class(es), the parent class(es)
            will be included in the reduced data model as well even though the parent class(es) are
            not in the desired classes set. This is to ensure that the reduced data model is
            consistent and complete.

        !!! note "Partial Reduction"
            This method does not perform checks if classes that are value types of desired classes
            properties are part of desired classes. If a class is not part of desired classes, but it
            is a value type of a property of a class that is part of desired classes, derived reduced
            rules will be marked as partial.

        !!! note "Validation"
            This method will attempt to validate the reduced rules with custom validations.
            If it fails, it will return a partial rules with a warning message, validated
            only with base Pydantic validators.
        """
        if self.rules.metadata.schema_ is not SchemaCompleteness.complete:
            raise ValueError("Rules are not complete cannot perform reduction!")
        class_as_dict = self.as_class_dict()
        class_parents_pairs = self.class_parent_pairs()
        defined_classes = self.defined_classes(consider_inheritance=True)

        possible_classes = defined_classes.intersection(desired_classes)
        impossible_classes = desired_classes - possible_classes

        # need to add all the parent classes of the desired classes to the possible classes
        parents: set[ClassEntity] = set()
        for class_ in possible_classes:
            parents = parents.union({parent for parent in get_inheritance_path(class_, class_parents_pairs)})
        possible_classes = possible_classes.union(parents)

        if not possible_classes:
            logging.error("None of the desired classes are defined in the data model!")
            raise ValueError("None of the desired classes are defined in the data model!")

        if impossible_classes:
            logging.warning(f"Could not find the following classes defined in the data model: {impossible_classes}")
            warnings.warn(
                f"Could not find the following classes defined in the data model: {impossible_classes}",
                stacklevel=2,
            )

        reduced_data_model: dict[str, Any] = {
            "metadata": self.rules.metadata.model_copy(),
            "prefixes": (self.rules.prefixes or {}).copy(),
            "classes": [],
            "properties": [],
        }

        logging.info(f"Reducing data model to only include the following classes: {possible_classes}")
        for class_ in possible_classes:
            reduced_data_model["classes"].append(class_as_dict[str(class_.suffix)])

        class_property_pairs = self.classes_with_properties(consider_inheritance=False)

        for class_, properties in class_property_pairs.items():
            if class_ in possible_classes:
                reduced_data_model["properties"].extend(properties)

        try:
            return type(self.rules)(**reduced_data_model)
        except ValidationError as e:
            warnings.warn(f"Reduced data model is not complete: {e}", stacklevel=2)
            reduced_data_model["metadata"].schema_ = SchemaCompleteness.partial
            return type(self.rules).model_construct(**reduced_data_model)



from typing import TypeAlias

from cognite.neat.rules.models import (
    AssetRules,
    DMSRules,
    DomainRules,
    InformationRules,
)

Rules: TypeAlias = DomainRules | InformationRules | DMSRules | AssetRules



from dataclasses import dataclass
from typing import Any

from cognite.neat.issues import NeatError, NeatWarning

__all__ = [
    "FailedAuthorizationError",
    "MissingDataModelError",
    "FailedConvertError",
    "InvalidClassWarning",
    "InvalidInstanceError",
]


@dataclass(frozen=True)
class FailedAuthorizationError(NeatError):
    description = "Missing authorization for {action}: {reason}"

    action: str
    reason: str

    def message(self) -> str:
        return self.description.format(action=self.action, reason=self.reason)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["action"] = self.action
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class MissingDataModelError(NeatError):
    description = "The data model with identifier {identifier} is missing: {reason}"
    fix = "Check the data model identifier and try again."

    identifier: str
    reason: str

    def message(self) -> str:
        return self.description.format(identifier=self.identifier, reason=self.reason)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["identifier"] = self.identifier
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class FailedConvertError(NeatError):
    description = "Failed to convert the {identifier} to {target_format}: {reason}"
    fix = "Check the error message and correct the rules."
    identifier: str
    target_format: str
    reason: str

    def message(self) -> str:
        return self.description.format(identifier=self.identifier, target_format=self.target_format, reason=self.reason)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["identifier"] = self.identifier
        output["targetFormat"] = self.target_format
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class InvalidClassWarning(NeatWarning):
    description = "The class {class_name} is invalid and will be skipped. {reason}"
    fix = "Check the error message and correct the class."

    class_name: str
    reason: str

    def message(self) -> str:
        return self.description.format(class_name=self.class_name, reason=self.reason)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["class_name"] = self.class_name
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class InvalidInstanceError(NeatError):
    description = "The {type_} with identifier {identifier} is invalid and will be skipped. {reason}"
    fix = "Check the error message and correct the instance."

    type_: str
    identifier: str
    reason: str

    def message(self) -> str:
        return self.description.format(type_=self.type_, identifier=self.identifier, reason=self.reason)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["type"] = self.type_
        output["identifier"] = self.identifier
        output["reason"] = self.reason
        return output



import re
from typing import cast

from rdflib import Graph, URIRef

from cognite.neat.rules.analysis import InformationAnalysis
from cognite.neat.rules.models._rdfpath import (
    Hop,
    RDFPath,
    SelfReferenceProperty,
    SingleProperty,
    Traversal,
)
from cognite.neat.rules.models.entities import ClassEntity
from cognite.neat.rules.models.information import InformationProperty, InformationRules
from cognite.neat.utils.collection_ import most_occurring_element

from ._shared import Triple, hop2property_path

_QUERY_TEMPLATE = """CONSTRUCT {{ {graph_template} }}
                     WHERE {{     {bind_instance_id}
                                  {graph_pattern}
                     }}"""


def build_construct_query(
    class_: ClassEntity,
    graph: Graph,
    rules: InformationRules,
    properties_optional: bool = True,
    instance_id: URIRef | None = None,
) -> str | None:
    """Builds a CONSTRUCT query for a given class and rules and optionally filters by class instances.

    Args:
        class_ : The class entity for which the query is generated.
        graph : The graph containing instances of classes.
        rules : The information rules to use for query generation.
        properties_optional : Whether to make all properties optional. Defaults to True.
        class_instances : List of class instances to filter by. Defaults to None (no filter, return all instances).

    Returns:
        str: CONSTRUCT query.

    !!! note "On CONSTRUCT Query"
        CONSTRUCT query is composed of two parts: graph template and graph pattern.
        Graph template is used the shape of instance acquired using graph pattern.
        This allows us to create a new graph with the new shape without actually modifying
        the original graph, or creating new instances.

        The CONSTRUCT query is far less forgiving than the SELECT query. It will not return
        anything if one of the properties that define the "shape" of the class instance is missing.
        This is the reason why there is an option to make all properties optional, so that
        the query will return all instances that have at least one property defined.
    """

    if (
        transformations := InformationAnalysis(rules)
        .class_property_pairs(only_rdfpath=True, consider_inheritance=True)
        .get(class_, None)
    ):
        templates, patterns = to_construct_triples(
            graph, list(transformations.values()), rules.prefixes, properties_optional
        )

        return _QUERY_TEMPLATE.format(
            bind_instance_id=(f"BIND(<{instance_id}> AS ?instance)" if instance_id else ""),
            graph_template="\n".join(triples2sparql_statement(templates)),
            graph_pattern="\n".join(triples2sparql_statement(patterns)),
        )

    else:
        return None


def to_construct_triples(
    graph: Graph,
    transformations: list[InformationProperty],
    prefixes: dict,
    properties_optional: bool = True,
) -> tuple[list[Triple], list[Triple]]:
    """Converts transformations of a class to CONSTRUCT triples which are used to generate CONSTRUCT query

    Args:
        graph: Graph containing instances of classes (used for property inference for hops)
        transformations : List of transformations to use to form triples
        prefixes : Dictionary of prefixes for namespaces
        properties_optional : Flag indicating if properties should be optional. Defaults to True.

    Returns:
        tuple: Tuple of triples that define graph template and graph pattern parts of CONSTRUCT query


    !!! note "Purely inherited transformations"
        Assumption that neat makes is that in case of purely inherited transformations
        we will type instance with class to which transformation belongs to.

        Otherwise we will type instance with class that is most occurring in non-inherited
        transformations.

    """
    # TODO: Add handling of UNIONs in rules

    templates = []
    patterns = []
    non_inherited_starting_rdf_types = []

    for transformation in transformations:
        traversal = cast(RDFPath, transformation.transformation).traversal

        # keeping track of starting rdf types of non-inherited transformations/properties
        if isinstance(traversal, Traversal) and not transformation.inherited:
            non_inherited_starting_rdf_types.append(traversal.class_.id)

        graph_template_triple = Triple(
            subject="?instance",
            predicate=f"{transformation.class_.prefix}:{transformation.property_}",
            object=f'?{re.sub(r"[^_a-zA-Z0-9/_]", "_", str(transformation.property_).lower())}',
            optional=False,
        )
        templates.append(graph_template_triple)

        # use case AllReferences: binding instance to certain rdf property
        if isinstance(traversal, SelfReferenceProperty):
            graph_pattern_triple = Triple(
                subject="BIND(?instance",
                predicate="AS",
                object=f"{graph_template_triple.object})",
                optional=False,
            )

        # use case SingleProperty: simple property traversal
        elif isinstance(traversal, SingleProperty):
            graph_pattern_triple = Triple(
                subject=graph_template_triple.subject,
                predicate=traversal.property.id,
                object=graph_template_triple.object,
                optional=(True if properties_optional else not transformation.is_mandatory),
            )

        # use case Hop: property traversal with multiple hops turned into property path
        # see: https://www.oxfordsemantic.tech/faqs/what-is-a-property-path
        elif isinstance(traversal, Hop):
            graph_pattern_triple = Triple(
                subject="?instance",
                predicate=hop2property_path(graph, traversal, prefixes),
                object=graph_template_triple.object,
                optional=(True if properties_optional else not transformation.is_mandatory),
            )

        # other type of rdfpaths are skipped
        else:
            continue

        patterns.append(graph_pattern_triple)

    # Add first triple for graph pattern stating type of object
    # we use most occurring here to pull out most occurring rdf type of the starting
    # node of the transformation, or the class itself to which the transformation is
    # defined for.
    # This is safeguard in case there are multiple classes in the graph pattern
    patterns.insert(
        0,
        Triple(
            subject="?instance",
            predicate="a",
            object=(
                most_occurring_element(non_inherited_starting_rdf_types)
                if non_inherited_starting_rdf_types
                else str(transformation.class_)
            ),
            optional=False,
        ),
    )

    return templates, patterns


def triples2sparql_statement(triples: list[Triple]):
    return [
        (
            f"OPTIONAL {{ {triple.subject} {triple.predicate} {triple.object} . }}"
            if triple.optional
            else f"{triple.subject} {triple.predicate} {triple.object} ."
        )
        for triple in triples
    ]



from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import Generic, TypeVar

from cognite.client import CogniteClient

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models import DMSRules, InformationRules, RoleTypes
from cognite.neat.utils.auxiliary import class_html_doc
from cognite.neat.utils.upload import UploadResult, UploadResultList

T_Export = TypeVar("T_Export")


class BaseExporter(ABC, Generic[T_Export]):
    _new_line = "\n"
    _encoding = "utf-8"

    @abstractmethod
    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def export(self, rules: Rules) -> T_Export:
        raise NotImplementedError

    def _convert_to_output_role(self, rules: Rules, output_role: RoleTypes | None = None) -> Rules:
        if rules.metadata.role is output_role or output_role is None:
            return rules
        elif output_role is RoleTypes.dms and isinstance(rules, InformationRules):
            return rules.as_dms_rules()
        elif output_role is RoleTypes.information and isinstance(rules, DMSRules):
            return rules.as_information_rules()
        else:
            raise NotImplementedError(f"Role {output_role} is not supported for {type(rules).__name__} rules")

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls, include_factory_methods=False)


class CDFExporter(BaseExporter[T_Export]):
    @abstractmethod
    def export_to_cdf_iterable(
        self, rules: Rules, client: CogniteClient, dry_run: bool = False
    ) -> Iterable[UploadResult]:
        raise NotImplementedError

    def export_to_cdf(self, rules: Rules, client: CogniteClient, dry_run: bool = False) -> UploadResultList:
        return UploadResultList(self.export_to_cdf_iterable(rules, client, dry_run))



import sys
import warnings
from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Sequence
from dataclasses import dataclass
from functools import total_ordering
from typing import Any, ClassVar, TypeVar
from warnings import WarningMessage

import pandas as pd

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
    from typing_extensions import Self
else:
    from typing import Self


@total_ordering
@dataclass(frozen=True)
class NeatIssue(ABC):
    description: ClassVar[str]
    fix: ClassVar[str]

    def message(self) -> str:
        """Return a human-readable message for the issue.

        This is the default implementation, which returns the description.
        It is recommended to override this method in subclasses with a more
        specific message.
        """
        return self.description

    @abstractmethod
    def dump(self) -> dict[str, Any]:
        """Return a dictionary representation of the issue."""
        raise NotImplementedError()

    def __lt__(self, other: "NeatIssue") -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.message()) < (type(other).__name__, other.message())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.message()) == (type(other).__name__, other.message())


@dataclass(frozen=True)
class NeatError(NeatIssue, ABC):
    def dump(self) -> dict[str, Any]:
        return {"errorType": type(self).__name__}

    def as_exception(self) -> ValueError:
        return ValueError(self.message())


@dataclass(frozen=True)
class NeatWarning(NeatIssue, ABC, UserWarning):
    def dump(self) -> dict[str, Any]:
        return {"warningType": type(self).__name__}

    @classmethod
    def from_warning(cls, warning: WarningMessage) -> "NeatWarning":
        return DefaultWarning.from_warning_message(warning)


@dataclass(frozen=True)
class DefaultWarning(NeatWarning):
    description = "A warning was raised during validation."
    fix = "No fix is available."

    warning: str | Warning
    category: type[Warning]
    source: str | None = None

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["msg"] = str(self.warning)
        output["category"] = self.category.__name__
        output["source"] = self.source
        return output

    @classmethod
    def from_warning_message(cls, warning: WarningMessage) -> NeatWarning:
        if isinstance(warning.message, NeatWarning):
            return warning.message

        return cls(
            warning=warning.message,
            category=warning.category,
            source=warning.source,
        )

    def message(self) -> str:
        return str(self.warning)


T_NeatIssue = TypeVar("T_NeatIssue", bound=NeatIssue)


class NeatIssueList(UserList[T_NeatIssue], ABC):
    def __init__(self, issues: Sequence[T_NeatIssue] | None = None, title: str | None = None):
        super().__init__(issues or [])
        self.title = title

    @property
    def errors(self) -> Self:
        return type(self)([issue for issue in self if isinstance(issue, NeatError)])  # type: ignore[misc]

    @property
    def has_errors(self) -> bool:
        return any(isinstance(issue, NeatError) for issue in self)

    @property
    def warnings(self) -> Self:
        return type(self)([issue for issue in self if isinstance(issue, NeatWarning)])  # type: ignore[misc]

    def as_errors(self) -> ExceptionGroup:
        return ExceptionGroup(
            "Operation failed",
            [ValueError(issue.message()) for issue in self if isinstance(issue, NeatError)],
        )

    def trigger_warnings(self) -> None:
        for warning in [issue for issue in self if isinstance(issue, NeatWarning)]:
            warnings.warn(warning, stacklevel=2)

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame([issue.dump() for issue in self])

    def _repr_html_(self) -> str | None:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]

    def as_exception(self) -> "MultiValueError":
        return MultiValueError(self.errors)


class MultiValueError(ValueError):
    """This is a container for multiple errors.

    It is used in the pydantic field_validator/model_validator to collect multiple errors, which
    can then be caught in a try-except block and returned as an IssueList.

    """

    def __init__(self, errors: Sequence[T_NeatIssue]):
        self.errors = list(errors)
